"""Defines a jax-compatible simulation."""
from __future__ import annotations

from typing import Tuple, Union, List, Dict
from multiprocessing import Pool

import pydantic.v1 as pd
import numpy as np

from jax.tree_util import register_pytree_node_class

from ....log import log
from ....components.base import cached_property, Tidy3dBaseModel
from ....components.monitor import FieldMonitor, PermittivityMonitor
from ....components.monitor import ModeMonitor, DiffractionMonitor, Monitor
from ....components.simulation import Simulation
from ....components.data.monitor_data import FieldData, PermittivityData
from ....components.structure import Structure
from ....components.types import Ax, annotate_type
from ....constants import HERTZ, SECOND
from ....exceptions import AdjointError

from .base import JaxObject
from .structure import JaxStructure
from .geometry import JaxPolySlab, JaxGeometryGroup


# bandwidth of adjoint source in units of freq0 if no `fwidth_adjoint`, and one output freq
FWIDTH_FACTOR = 1.0 / 10

# bandwidth of adjoint sources in units of the minimum difference between output frequencies
FWIDTH_FACTOR_MULTIFREQ = 0.1

# the adjoint run time is RUN_TIME_FACTOR / fwidth
RUN_TIME_FACTOR = 100

# how many processors to use for server and client side adjoint
NUM_PROC_LOCAL = 1

# number of input structures before it errors
MAX_NUM_INPUT_STRUCTURES = 400


class JaxInfo(Tidy3dBaseModel):
    """Class to store information when converting between jax and tidy3d."""

    num_input_structures: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Input Structures",
        description="Number of input structures in the original JaxSimulation.",
    )

    num_output_monitors: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Output Monitors",
        description="Number of output monitors in the original JaxSimulation.",
    )

    num_grad_monitors: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Gradient Monitors",
        description="Number of gradient monitors in the original JaxSimulation.",
    )

    num_grad_eps_monitors: pd.NonNegativeInt = pd.Field(
        ...,
        title="Number of Permittivity Monitors",
        description="Number of permittivity monitors in the original JaxSimulation.",
    )

    fwidth_adjoint: float = pd.Field(
        None,
        title="Adjoint Frequency Width",
        description="Custom frequency width of the original JaxSimulation.",
        units=HERTZ,
    )

    run_time_adjoint: float = pd.Field(
        None,
        title="Adjoint Run Time",
        description="Custom run time of the original JaxSimulation.",
        units=SECOND,
    )


@register_pytree_node_class
class JaxSimulation(Simulation, JaxObject):
    """A :class:`.Simulation` registered with jax."""

    input_structures: Tuple[JaxStructure, ...] = pd.Field(
        (),
        title="Input Structures",
        description="Tuple of jax-compatible structures"
        " that may depend on differentiable parameters.",
        jax_field=True,
    )

    output_monitors: Tuple[
        annotate_type(Union[DiffractionMonitor, FieldMonitor, ModeMonitor]), ...
    ] = pd.Field(
        (),
        title="Output Monitors",
        description="Tuple of monitors whose data the differentiable output depends on.",
    )

    grad_monitors: Tuple[FieldMonitor, ...] = pd.Field(
        (),
        title="Gradient Field Monitors",
        description="Tuple of monitors used for storing fields, used internally for gradients.",
    )

    grad_eps_monitors: Tuple[PermittivityMonitor, ...] = pd.Field(
        (),
        title="Gradient Permittivity Monitors",
        description="Tuple of monitors used for storing epsilon, used internally for gradients.",
    )

    fwidth_adjoint: pd.PositiveFloat = pd.Field(
        None,
        title="Adjoint Frequency Width",
        description="Custom frequency width to use for ``source_time`` of adjoint sources. "
        "If not supplied or ``None``, uses the average fwidth of the original simulation's sources.",
        units=HERTZ,
    )

    run_time_adjoint: pd.PositiveFloat = pd.Field(
        None,
        title="Adjoint Run Time",
        description="Custom ``run_time`` to use for adjoint simulation. "
        "If not supplied or ``None``, uses a factor times the adjoint source ``fwidth``.",
        units=SECOND,
    )

    @pd.validator("output_monitors", always=True)
    def _output_monitors_colocate_false(cls, val):
        """Make sure server-side colocation is off."""
        new_vals = []
        for mnt in val:
            if mnt.colocate:
                log.warning(
                    "Output field monitors in the adjoint plugin require explicitly setting "
                    "'colocate=False'. Setting 'colocate=False' in monitor '{mnt.name}'. ",
                    "Use 'SimulationData.at_boundaries' after the solver run to automatically "
                    "colocate the fields to the grid boundaries, or 'MonitorData.colocate' "
                    "if colocating to custom coordinates.",
                )
                mnt = mnt.updated_copy(colocate=False)
            new_vals.append(mnt)
        return new_vals

    @pd.validator("subpixel", always=True)
    def _subpixel_is_on(cls, val):
        """Assert subpixel is on."""
        if not val:
            raise AdjointError("'JaxSimulation.subpixel' must be 'True' to use adjoint plugin.")
        return val

    @pd.validator("input_structures", always=True)
    def _restrict_input_structures(cls, val):
        """Restrict number of input structures."""
        num_input_structures = len(val)
        if num_input_structures > MAX_NUM_INPUT_STRUCTURES:
            raise AdjointError(
                "For performance, adjoint plugin restricts the number of input structures to "
                f"{MAX_NUM_INPUT_STRUCTURES}. Found {num_input_structures}."
            )

        return val

    @pd.validator("input_structures", always=True)
    def _warn_overlap(cls, val, values):
        """Print appropriate warning if structures intersect in ways that cause gradient error."""

        input_structures = list(val)
        structures = list(values.get("structures"))

        # if the center and size of all structure geometries do not contain all numbers, skip check
        for struct in input_structures:
            geometry = struct.geometry
            size_all_floats = all(isinstance(s, (float, int)) for s in geometry.bound_size)
            cent_all_floats = all(isinstance(c, (float, int)) for c in geometry.bound_center)
            if not (size_all_floats and cent_all_floats):
                return val

        with log as consolidated_logger:
            # check intersections with other input_structures
            for i, in_struct_i in enumerate(input_structures):
                geometry_i = in_struct_i.geometry
                for j in range(i + 1, len(input_structures)):
                    if geometry_i.intersects(input_structures[j].geometry):
                        consolidated_logger.warning(
                            f"'JaxSimulation.input_structures[{i}]' overlaps or touches "
                            f"'JaxSimulation.input_structures[{j}]'. Geometric gradients for "
                            "overlapping input structures may contain errors."
                        )

            # check JaxPolySlab intersections with background structures
            for i, in_struct_i in enumerate(input_structures):
                geometry_i = in_struct_i.geometry
                if not isinstance(geometry_i, JaxPolySlab):
                    continue
                for j, struct_j in enumerate(structures):
                    if geometry_i.intersects(struct_j.geometry):
                        consolidated_logger.warning(
                            f"'JaxPolySlab'-containing 'JaxSimulation.input_structures[{i}]' "
                            f"intersects with 'JaxSimulation.structures[{j}]'. Note that in this "
                            "version of the adjoint plugin, there may be errors in the gradient "
                            "when 'JaxPolySlab' intersects with background structures."
                        )

        return val

    @pd.validator("output_monitors", always=True)
    def _warn_if_colocate(cls, val):
        """warn if any colocate=True in output FieldMonitors."""
        for index, mnt in enumerate(val):
            if isinstance(mnt, FieldMonitor):
                if mnt.colocate:
                    log.warning(
                        f"'FieldMonitor' at 'JaxSimulation.output_monitors[{index}]' "
                        "has 'colocate=True', "
                        "this may lead to decreased accuracy in adjoint gradient."
                    )
                    return val
        return val

    @staticmethod
    def get_freqs_adjoint(output_monitors: List[Monitor]) -> List[float]:
        """Return sorted list of unique frequencies stripped from a collection of monitors."""

        if len(output_monitors) == 0:
            raise AdjointError("Can't get adjoint frequency as no output monitors present.")

        output_freqs = []
        for mnt in output_monitors:
            for freq in mnt.freqs:
                output_freqs.append(freq)

        return np.unique(output_freqs).tolist()

    @cached_property
    def freqs_adjoint(self) -> List[float]:
        """Return sorted list of frequencies stripped from the output monitors."""
        return self.get_freqs_adjoint(output_monitors=self.output_monitors)

    @cached_property
    def _is_multi_freq(self) -> bool:
        """Does this simulation have a multi-frequency output?"""
        return len(self.freqs_adjoint) > 1

    @cached_property
    def _min_delta_freq(self) -> float:
        """Minimum spacing between output_frequencies (Hz)."""

        if not self._is_multi_freq:
            return None

        delta_freqs = np.abs(np.diff(np.sort(np.array(self.freqs_adjoint))))
        return np.min(delta_freqs)

    @cached_property
    def _fwidth_adjoint(self) -> float:
        """Frequency width to use for adjoint source, user-defined or the average of the sources."""

        # if user-specified, use that
        if self.fwidth_adjoint is not None:
            return self.fwidth_adjoint

        freqs_adjoint = self.freqs_adjoint

        # multiple output frequency case
        if self._is_multi_freq:
            return FWIDTH_FACTOR_MULTIFREQ * self._min_delta_freq

        # otherwise, grab from sources and output monitors
        num_sources = len(self.sources)  # should be 0 for adjoint already but worth checking

        # if no sources, just use a constant factor times the mean adjoint frequency
        if num_sources == 0:
            return FWIDTH_FACTOR * np.mean(freqs_adjoint)

        # if more than one forward source, use their maximum
        if num_sources > 1:
            log.warning(f"{num_sources} sources, using their maximum 'fwidth' for adjoint source.")

        fwidths = [src.source_time.fwidth for src in self.sources]
        return np.max(fwidths)

    @cached_property
    def _run_time_adjoint(self: float) -> float:
        """Return the run time of the adjoint simulation as a function of its fwidth."""

        if self.run_time_adjoint is not None:
            return self.run_time_adjoint

        run_time_adjoint = RUN_TIME_FACTOR / self._fwidth_adjoint

        if self._is_multi_freq:

            log.warning(
                f"{len(self.freqs_adjoint)} unique frequencies detected in the output monitors "
                f"with a minimum spacing of {self._min_delta_freq:.3e} (Hz). "
                f"Setting the 'fwidth' of the adjoint sources to {FWIDTH_FACTOR_MULTIFREQ} times "
                f"this value = {self._fwidth_adjoint:.3e} (Hz) to avoid spectral overlap. "
                "To account for this, the corresponding 'run_time' in the adjoint simulation is "
                f"will be set to {run_time_adjoint:3e} "
                f"compared to {self.run_time:3e} in the forward simulation. "
                "If the adjoint 'run_time' is large due to small frequency spacing, "
                "it could be better to instead run one simulation per frequency, "
                "which can be done in parallel using 'tidy3d.plugins.adjoint.web.run_async'."
            )

        return run_time_adjoint

    def to_simulation(self) -> Tuple[Simulation, JaxInfo]:
        """Convert :class:`.JaxSimulation` instance to :class:`.Simulation` with an info dict."""

        sim_dict = self.dict(
            exclude={
                "type",
                "structures",
                "monitors",
                "output_monitors",
                "grad_monitors",
                "grad_eps_monitors",
                "input_structures",
                "fwidth_adjoint",
                "run_time_adjoint",
            }
        )
        sim = Simulation.parse_obj(sim_dict)

        # put all structures and monitors in one list
        all_structures = list(self.structures) + [js.to_structure() for js in self.input_structures]
        all_monitors = (
            list(self.monitors)
            + list(self.output_monitors)
            + list(self.grad_monitors)
            + list(self.grad_eps_monitors)
        )

        sim = sim.updated_copy(structures=all_structures, monitors=all_monitors)

        # information about the state of the original JaxSimulation to stash for reconstruction
        jax_info = JaxInfo(
            num_input_structures=len(self.input_structures),
            num_output_monitors=len(self.output_monitors),
            num_grad_monitors=len(self.grad_monitors),
            num_grad_eps_monitors=len(self.grad_eps_monitors),
            fwidth_adjoint=self.fwidth_adjoint,
            run_time_adjoint=self.run_time_adjoint,
        )

        return sim, jax_info

    def plot(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        source_alpha: float = None,
        monitor_alpha: float = None,
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
        **patch_kwargs,
    ) -> Ax:
        """Wrapper around regular :class:`.Simulation` structure plotting."""
        sim, _ = self.to_simulation()
        return sim.plot(
            x=x,
            y=y,
            z=z,
            ax=ax,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            hlim=hlim,
            vlim=vlim,
            **patch_kwargs,
        )

    def plot_eps(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        freq: float = None,
        alpha: float = None,
        source_alpha: float = None,
        monitor_alpha: float = None,
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
        ax: Ax = None,
    ) -> Ax:
        """Wrapper around regular :class:`.Simulation` permittivity plotting."""
        sim, _ = self.to_simulation()
        return sim.plot_eps(
            x=x,
            y=y,
            z=z,
            ax=ax,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            hlim=hlim,
            vlim=vlim,
        )

    def plot_structures(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        sim, _ = self.to_simulation()
        return sim.plot_structures(
            x=x,
            y=y,
            z=z,
            ax=ax,
            hlim=hlim,
            vlim=vlim,
        )

    def plot_structures_eps(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        freq: float = None,
        alpha: float = None,
        cbar: bool = True,
        reverse: bool = False,
        ax: Ax = None,
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
    ) -> Ax:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z coordinate.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        reverse : bool = False
            If ``False``, the highest permittivity is plotted in black.
            If ``True``, it is plotteed in white (suitable for black backgrounds).
        cbar : bool = True
            Whether to plot a colorbar for the relative permittivity.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        sim, _ = self.to_simulation()
        return sim.plot_structures_eps(
            x=x,
            y=y,
            z=z,
            freq=freq,
            alpha=alpha,
            cbar=cbar,
            reverse=reverse,
            ax=ax,
            hlim=hlim,
            vlim=vlim,
        )

    def __eq__(self, other: JaxSimulation) -> bool:
        """Are two JaxSimulation objects equal?"""
        return self.to_simulation()[0] == other.to_simulation()[0]

    @classmethod
    def split_monitors(cls, monitors: List[Monitor], jax_info: JaxInfo) -> Dict[str, Monitor]:
        """Split monitors into user and adjoint required based on jax info."""

        all_monitors = list(monitors)

        # grab or compute the number of type of monitor
        num_grad_monitors = jax_info.num_grad_monitors
        num_grad_eps_monitors = jax_info.num_grad_eps_monitors
        num_output_monitors = jax_info.num_output_monitors
        num_total_monitors = len(all_monitors)
        num_mnts = (
            num_total_monitors - num_grad_monitors - num_output_monitors - num_grad_eps_monitors
        )

        # split the monitor list based on these numbers
        monitors = all_monitors[:num_mnts]
        output_monitors = all_monitors[num_mnts : num_mnts + num_output_monitors]
        grad_monitors = all_monitors[
            num_mnts + num_output_monitors : num_mnts + num_output_monitors + num_grad_monitors
        ]
        grad_eps_monitors = all_monitors[num_mnts + num_output_monitors + num_grad_monitors :]

        # load into a dictionary
        return dict(
            monitors=monitors,
            output_monitors=output_monitors,
            grad_monitors=grad_monitors,
            grad_eps_monitors=grad_eps_monitors,
        )

    @classmethod
    def split_structures(
        cls, structures: List[Structure], jax_info: JaxInfo
    ) -> Dict[str, Structure]:
        """Split structures into regular and input based on jax info."""

        all_structures = list(structures)

        # get numbers of regular and input structures
        num_input_structures = jax_info.num_input_structures
        num_structs = len(structures) - num_input_structures

        # split the list based on these numbers
        structures = all_structures[:num_structs]
        input_structures = [JaxStructure.from_structure(s) for s in all_structures[num_structs:]]

        # return a dictionary containing these split structures
        return dict(structures=structures, input_structures=input_structures)

    @classmethod
    def from_simulation(cls, simulation: Simulation, jax_info: JaxInfo) -> JaxSimulation:
        """Convert :class:`.Simulation` to :class:`.JaxSimulation` with extra info."""

        sim_dict = simulation.dict(exclude={"type", "structures", "monitors"})

        # split structures and monitors into their respective fields for JaxSimulation
        structures = cls.split_structures(structures=simulation.structures, jax_info=jax_info)
        monitors = cls.split_monitors(monitors=simulation.monitors, jax_info=jax_info)

        # update the dictionary with these and the adjoint fwidth
        sim_dict.update(**structures)
        sim_dict.update(**monitors)
        sim_dict.update(
            dict(
                fwidth_adjoint=jax_info.fwidth_adjoint,
                run_time_adjoint=jax_info.run_time_adjoint,
            )
        )

        # load JaxSimulation from the dictionary
        return cls.parse_obj(sim_dict)

    @classmethod
    def make_sim_fwd(cls, simulation: Simulation, jax_info: JaxInfo) -> Tuple[Simulation, JaxInfo]:
        """Make the forward :class:`.JaxSimulation` from the supplied :class:`.Simulation`."""

        mnt_dict = JaxSimulation.split_monitors(monitors=simulation.monitors, jax_info=jax_info)
        structure_dict = JaxSimulation.split_structures(
            structures=simulation.structures, jax_info=jax_info
        )
        output_monitors = mnt_dict["output_monitors"]
        input_structures = structure_dict["input_structures"]
        grad_mnt_dict = cls.get_grad_monitors(
            input_structures=input_structures,
            freqs_adjoint=cls.get_freqs_adjoint(output_monitors=output_monitors),
        )

        grad_mnts = grad_mnt_dict["grad_monitors"]
        grad_eps_mnts = grad_mnt_dict["grad_eps_monitors"]

        full_monitors = list(simulation.monitors) + grad_mnts + grad_eps_mnts

        # jax_sim_fwd = jax_sim.updated_copy(**grad_mnts)
        # sim_fwd, jax_info = jax_sim_fwd.to_simulation()

        sim_fwd = simulation.updated_copy(monitors=full_monitors)
        jax_info = jax_info.updated_copy(
            num_grad_monitors=len(grad_mnts),
            num_grad_eps_monitors=len(grad_eps_mnts),
        )

        # cls.split_monitors(monitors=simulation.monitors, jax_info=jax_info)
        # sim_fwd = simulation.updated_copy()

        return sim_fwd, jax_info

    def to_simulation_fwd(self) -> Tuple[Simulation, JaxInfo, JaxInfo]:
        """Like ``to_simulation()`` but the gradient monitors are included."""
        simulation, jax_info = self.to_simulation()
        sim_fwd, jax_info_fwd = self.make_sim_fwd(simulation=simulation, jax_info=jax_info)
        return sim_fwd, jax_info_fwd, jax_info

    @staticmethod
    def get_grad_monitors(
        input_structures: List[Structure], freqs_adjoint: List[float], include_eps_mnts: bool = True
    ) -> dict:
        """Return dictionary of gradient monitors for simulation."""
        grad_mnts = []
        grad_eps_mnts = []
        for index, structure in enumerate(input_structures):
            grad_mnt, grad_eps_mnt = structure.make_grad_monitors(
                freqs=freqs_adjoint, name=f"grad_mnt_{index}"
            )
            grad_mnts.append(grad_mnt)
            if include_eps_mnts:
                grad_eps_mnts.append(grad_eps_mnt)
        return dict(grad_monitors=grad_mnts, grad_eps_monitors=grad_eps_mnts)

    def _store_vjp_structure(
        self,
        structure: JaxStructure,
        fld_fwd: FieldData,
        fld_adj: FieldData,
        eps_data: PermittivityData,
        num_proc: int = NUM_PROC_LOCAL,
    ) -> JaxStructure:
        """Store the vjp for a single structure."""

        freq_max = float(max(eps_data.eps_xx.coords["f"]))
        eps_out = self.medium.eps_model(frequency=freq_max)
        return structure.store_vjp(
            grad_data_fwd=fld_fwd,
            grad_data_adj=fld_adj,
            grad_data_eps=eps_data,
            sim_bounds=self.bounds,
            eps_out=eps_out,
            num_proc=num_proc,
        )

    def store_vjp(
        self,
        grad_data_fwd: Tuple[FieldData],
        grad_data_adj: Tuple[FieldData],
        grad_eps_data: Tuple[PermittivityData],
        num_proc: int = NUM_PROC_LOCAL,
    ) -> JaxSimulation:
        """Store the vjp w.r.t. each input_structure as a sim using fwd and adj grad_data."""

        # if num_proc supplied and greater than 1, run parallel
        if num_proc is not None and num_proc > 1:
            return self.store_vjp_parallel(
                grad_data_fwd=grad_data_fwd,
                grad_data_adj=grad_data_adj,
                grad_eps_data=grad_eps_data,
                num_proc=num_proc,
            )

        # otherwise, call regular sequential one
        return self.store_vjp_sequential(
            grad_data_fwd=grad_data_fwd, grad_data_adj=grad_data_adj, grad_eps_data=grad_eps_data
        )

    def store_vjp_sequential(
        self,
        grad_data_fwd: Tuple[FieldData],
        grad_data_adj: Tuple[FieldData],
        grad_eps_data: Tuple[PermittivityData],
    ) -> JaxSimulation:
        """Store the vjp w.r.t. each input_structure without multiprocessing."""
        map_args = [self.input_structures, grad_data_fwd, grad_data_adj, grad_eps_data]
        input_structures_vjp = list(map(self._store_vjp_structure, *map_args))

        return self.copy(
            update=dict(
                input_structures=input_structures_vjp, grad_monitors=(), grad_eps_monitors=()
            )
        )

    def store_vjp_parallel(
        self,
        grad_data_fwd: Tuple[FieldData],
        grad_data_adj: Tuple[FieldData],
        grad_eps_data: Tuple[PermittivityData],
        num_proc: int,
    ) -> JaxSimulation:
        """Store the vjp w.r.t. each input_structure as a sim using fwd and adj grad_data, and
        parallel processing over ``num_proc`` processes."""

        # Indexing into structures which use internal parallelization, and those which don't.
        # For the latter, simple parallelization over the list will be used.
        internal_par_structs = [JaxGeometryGroup]

        # Parallelize polyslabs internally or externally depending on total number
        polyslabs = [struct for struct in self.input_structures if isinstance(struct, JaxPolySlab)]
        if len(polyslabs) < num_proc:
            internal_par_structs += [JaxPolySlab]

        inds_par_internal, inds_par_external = [], []
        for index, structure in enumerate(self.input_structures):
            if isinstance(structure.geometry, tuple(internal_par_structs)):
                inds_par_internal.append(index)
            else:
                inds_par_external.append(index)

        def make_args(indexes, num_proc_internal):
            """Make the arguments to map over selecting over a set of structure ``indexes``."""
            structures = [self.input_structures[index] for index in indexes]
            data_fwd = [grad_data_fwd[index] for index in indexes]
            data_bwd = [grad_data_adj[index] for index in indexes]
            eps_data = [grad_eps_data[index] for index in indexes]
            num_proc = [num_proc_internal] * len(indexes)

            return (structures, data_fwd, data_bwd, eps_data, num_proc)

        # Get vjps for structures that parallelize internally using simple map
        args_par_internal = make_args(inds_par_internal, num_proc_internal=num_proc)
        vjps_par_internal = list(map(self._store_vjp_structure, *args_par_internal))

        # Get vjps for structures where we parallelize directly here
        args_par_external = make_args(inds_par_external, num_proc_internal=NUM_PROC_LOCAL)
        with Pool(num_proc) as pool:
            vjps_par_external = list(
                pool.starmap(self._store_vjp_structure, zip(*args_par_external))
            )

        # Reshuffle the two lists back in the correct order
        vjps_all = list(vjps_par_internal) + list(vjps_par_external)
        input_structures_vjp = [None] * len(self.input_structures)
        for index, vjp in zip(inds_par_internal + inds_par_external, vjps_all):
            input_structures_vjp[index] = vjp

        return self.copy(
            update=dict(
                input_structures=input_structures_vjp, grad_monitors=(), grad_eps_monitors=()
            )
        )



"""Monitor level data, store the DataArrays associated with a single heat monitor."""
from __future__ import annotations
from typing import Union, Tuple

from abc import ABC

import numpy as np
import pydantic.v1 as pd

from ..monitor import TemperatureMonitor, HeatMonitorType
from ...base_sim.data.monitor_data import AbstractMonitorData
from ...data.data_array import SpatialDataArray
from ...types import ScalarSymmetry, Coordinate
from ....constants import KELVIN


class HeatMonitorData(AbstractMonitorData, ABC):
    """Abstract base class of objects that store data pertaining to a single :class:`HeatMonitor`."""

    monitor: HeatMonitorType = pd.Field(
        ...,
        title="Monitor",
        description="Monitor associated with the data.",
    )

    symmetry: Tuple[ScalarSymmetry, ScalarSymmetry, ScalarSymmetry] = pd.Field(
        (0, 0, 0),
        title="Symmetry",
        description="Symmetry of the original simulation in x, y, and z.",
    )

    symmetry_center: Coordinate = pd.Field(
        (0, 0, 0),
        title="Symmetry Center",
        description="Symmetry center of the original simulation in x, y, and z.",
    )

    @property
    def symmetry_expanded_copy(self) -> HeatMonitorData:
        """Return copy of self with symmetry applied."""
        return self.copy()


class TemperatureData(HeatMonitorData):
    """Data associated with a :class:`TemperatureMonitor`: spatial temperature field.

    Example
    -------
    >>> from tidy3d import TemperatureMonitor, SpatialDataArray
    >>> temp_data = SpatialDataArray(
    ...     np.ones((2, 3, 4)), coords={"x": [0, 1], "y": [0, 1, 2], "z": [0, 1, 2, 3]}
    ... )
    >>> temp_mnt = TemperatureMonitor(size=(1, 2, 3), name="temperature")
    >>> temp_mnt_data = TemperatureData(
    ...     monitor=temp_mnt, temperature=temp_data, symmetry=(0, 1, 0), symmetry_center=(0, 0, 0)
    ... )
    >>> temp_mnt_data_expanded = temp_mnt_data.symmetry_expanded_copy
    """

    monitor: TemperatureMonitor = pd.Field(
        ..., title="Monitor", description="Temperature monitor associated with the data."
    )

    temperature: SpatialDataArray = pd.Field(
        ..., title="Temperature", description="Spatial temperature field.", units=KELVIN
    )

    @property
    def symmetry_expanded_copy(self) -> TemperatureData:
        """Return copy of self with symmetry applied."""

        if all(sym == 0 for sym in self.symmetry):
            return self.copy()

        coords = list(self.temperature.coords.values())
        data = np.array(self.temperature.data)

        for dim in range(3):
            if self.symmetry[dim] == 1:

                sym_center = self.symmetry_center[dim]

                if sym_center == coords[dim].data[0]:
                    num_duplicates = 1
                else:
                    num_duplicates = 0

                shape = np.array(np.shape(data))
                old_len = shape[dim]
                shape[dim] = 2 * old_len - num_duplicates

                ind_left = [slice(shape[0]), slice(shape[1]), slice(shape[2])]
                ind_right = [slice(shape[0]), slice(shape[1]), slice(shape[2])]

                ind_left[dim] = slice(old_len - 1, None, -1)
                ind_right[dim] = slice(old_len - num_duplicates, None)

                new_data = np.zeros(shape)

                new_data[ind_left[0], ind_left[1], ind_left[2]] = data
                new_data[ind_right[0], ind_right[1], ind_right[2]] = data

                new_coords = np.zeros(shape[dim])
                new_coords[old_len - num_duplicates :] = coords[dim]
                new_coords[old_len - 1 :: -1] = 2 * sym_center - coords[dim]

                coords[dim] = new_coords
                data = new_data

        coords_dict = dict(zip("xyz", coords))
        new_temperature = SpatialDataArray(data, coords=coords_dict)

        return self.updated_copy(temperature=new_temperature, symmetry=(0, 0, 0))


HeatMonitorDataType = Union[TemperatureData]



from pathlib import Path


def _remove_spaces(line: str) -> str:
    """
    Removes spaces from line, except spaces found between quotes. Also converts " into ' for ease of further parsing
    """
    quotes = False
    comment = False
    new_line = ""
    for c in line:
        if c == " " and not quotes:
            if comment:
                new_line += c
        else:
            if c == "'" or c == '"':
                quotes = not quotes
                c = "'"
            elif c == "#":
                comment = True
            new_line += c
    return new_line


def _is_int(number_str: str) -> bool:
    try:
        int(number_str)
        return True
    except ValueError:
        return False


def _is_float(number_str: str) -> bool:
    try:
        float(number_str)
        return True
    except ValueError:
        return False


def _clean(sci_str: str) -> str:
    number, power = sci_str.split("e")
    while number[-1] == "0" or number[-1] == ".":
        if len(number) == 1:
            break
        number = number[:-1]
    while power[1] == "0" and len(power) > 2:
        power = power[0] + power[2:]
    if power[0] == "+":
        power = power[1:]
    if power == "0":
        return number
    if "." not in number and power == "1":
        return number + "0"
    return number + "e" + power


def _to_um(array: list) -> list:
    output_array = []
    for number_str in array:
        if number_str == "None":
            output_array.append("None")
        elif number_str == "td.inf" or number_str == "-td.inf":
            output_array.append(number_str)
        elif "e" in number_str:  # potentially scientific notation
            number_power = number_str.split("e")
            if len(number_power) != 2:  # number_str is a variable name with an 'e'
                output_array.append(number_str + "*1e6")
            else:
                number, power = number_power
                if _is_int(number):
                    if _is_int(power):
                        value = int(number) * 10 ** (int(power)) * 1e6
                        value_str = _clean(f"{value:e}")
                        output_array.append(value_str)
                    elif _is_float(power):
                        value = int(number) * 10 ** (float(power)) * 1e6
                        value_str = _clean(f"{value:e}")
                        output_array.append(value_str)
                    else:  # number_str something else
                        output_array.append(number_str + "*1e6")
                elif _is_float(number):
                    if _is_int(power):
                        value = float(number) * 10 ** (int(power)) * 1e6
                        value_str = _clean(f"{value:e}")
                        output_array.append(value_str)
                    elif _is_float(power):
                        value = float(number) * 10 ** (float(power)) * 1e6
                        value_str = _clean(f"{value:e}")
                        output_array.append(value_str)
                    else:  # number_str is something else
                        output_array.append(number_str + "*1e6")
                else:
                    output_array.append(number_str + "*1e6")
        elif _is_int(number_str):
            value_str = _clean(f"{int(number_str):e}")
            output_array.append(value_str)
        elif _is_float(number_str):
            value_str = _clean(f"{float(number_str):e}")
            output_array.append(value_str)
        else:  # dealing with a string representing a variable
            output_array.append(number_str + "*1e6")
    return output_array


def _is_declaration(Line: str) -> bool:
    comment = False
    for i in range(0, len(Line)):
        if Line[i] == "#":
            comment = True
        if Line[i] == "=" and not comment:
            if "[" in Line[i] and "]" in Line[i]:  # change array syntax
                l_bracket = Line[i].find("[")
                new_array_string = "["
                array = Line[l_bracket + 1 :].split(";")
                for a in array:
                    new_array_string += "(" + a + "),"
                r_bracket = new_array_string.find("]")
                new_array_string = (
                    new_array_string[:r_bracket] + ")" + new_array_string[r_bracket:-1]
                )
                Line[i] = Line[i][:l_bracket] + new_array_string
            return True
    return False


def _add_to_variable_dict(line: str, variable_dict: dict):
    """
    Takes a string which describes a variable assignment and adds the variable assignment to the variable dictionary. Supported for strings, floats, booleans, and other variables.

    E.g. "string_var='str'" adds 'string_var': 'str' to the dictionary
    """
    variable_name = ""
    var = False
    for i in range(0, len(line)):
        if line[i] != "=" and not var:
            variable_name += line[i]
        else:
            if line[i] == "=":
                var = True
            else:
                if line[i] == "'":
                    # if variable represents a string
                    variable_dict[variable_name] = line[i + 1 :].split("'")[0]
                elif line[i : i + 4] == "True":
                    variable_dict[variable_name] = True
                elif line[i : i + 5] == "False":
                    variable_dict[variable_name] = False
                elif line[i] == "[":
                    return  # is an array
                else:
                    try:  # if variable represents a float
                        float(line[i:].split(";")[0])
                        variable_dict[variable_name] = float(line[i:].split(";")[0])
                    except ValueError:
                        # if variable represents another variable
                        variable_dict[variable_name] = variable_dict[line[i:].split(";")[0]]
                return


def _addrect(Lines, i: int, rect: int) -> str:
    # https://optics.ansys.com/hc/en-us/articles/360034901473-Rectangle-Simulation-Object
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    x_min, x_max = "0", "0"
    y_min, y_max = "0", "0"
    z_min, z_max = "0", "0"
    name = "rect_" + str(rect)
    medium = ""
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:10] == "x_min":
            x_min = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "x_max":
            x_max = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "y_min":
            y_min = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "y_max":
            y_max = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "z_min":
            z_min = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:10] == "z_max":
            z_max = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:10] == "index":
            index = Lines[i][12:].split(";")[0][:-1]
            medium = "td.Medium(permittivity=" + index + "**2)"
        elif Lines[i][5:13] == "material":
            medium = Lines[i][15:].split(";")[0][:-1] + " # NOTE: check material library"
        i += 1
        if i == len(Lines):
            break
    rect_string = name.replace(" ", "_") + " = td.Structure(\n"
    rect_string += "\tgeometry=td.Box(\n"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    rect_string += "\t\tcenter=(" + x + "," + y + "," + z + "),\n"
    rect_string += "\t\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    rect_string += "\t),\n"
    rect_string += "\tmedium=" + medium + ",\n"
    rect_string += "\tname='" + name + "',\n"
    rect_string += ")\n"
    return rect_string, name.replace(" ", "_")


def _addsphere(Lines, i: int, sph: int):
    # https://optics.ansys.com/hc/en-us/articles/360034901553-Sphere-Simulation-Object
    x, y, z = "0", "0", "0"
    radius = " # NOTE: Please specify"
    name = "sphere_" + str(sph)
    medium = "td.Medium()"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "radius":
            radius = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:10] == "index":
            index = Lines[i][12:].split(";")[0][:-1]
            medium = "td.Medium(permittivity=" + index + "**2)"
        elif Lines[i][5:13] == "material":
            medium = Lines[i][15:].split(";")[0][:-1] + " # NOTE: check material library"
        elif Lines[i][5:19] == "make ellipsoid":
            pass  # for now
        i += 1
        if i == len(Lines):
            break
    sphere_string = name.replace(" ", "_") + " = td.Structure(\n"
    sphere_string += "\tgeometry=td.Sphere(\n"
    x, y, z, radius = _to_um([x, y, z, radius])
    sphere_string += "\t\tcenter=(" + x + "," + y + "," + z + "),\n"
    sphere_string += "\t\tradius=" + radius + ",\n"
    sphere_string += "\t),\n"
    sphere_string += "\tmedium=" + medium + ",\n"
    sphere_string += "\tname='" + name + "',\n"
    sphere_string += ")\n"
    return sphere_string, name.replace(" ", "_")


def _addcircle(Lines, i: int, cyl: int):
    # https://optics.ansys.com/hc/en-us/articles/360034901513-Circle-Simulation-Object
    x, y, z = "0", "0", "0"
    z_min, z_max = "-td.inf/2", "td.inf/2"
    z_span = z_max + "-" + z_min
    name = "cylinder_" + str(cyl)
    radius = " # NOTE: please specify"
    medium = "td.Medium()"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:10] == "z min":
            z_min = Lines[i][12:].split(";")[0][:-1]
            z_span = z_max + "-" + z_min
        elif Lines[i][5:10] == "z max":
            z_max = Lines[i][12:].split(";")[0][:-1]
            z_span = z_max + "-" + z_min
        elif Lines[i][5:11] == "z span":
            z_span = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "radius":
            radius = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:10] == "index":
            index = Lines[i][12:].split(";")[0][:-1]
            medium = "td.Medium(permittivity=" + index + "**2)"
        elif Lines[i][5:13] == "material":
            medium = Lines[i][15:].split(";")[0][:-1] + " # NOTE: check material library"
        i += 1
        if i == len(Lines):
            break
    cylinder_string = name.replace(" ", "_") + " = td.Structure(\n"
    cylinder_string += "\tgeometry=td.Cylinder(\n"
    x, y, z, radius, z_span = _to_um([x, y, z, radius, z_span])
    cylinder_string += "\t\tcenter=(" + x + "," + y + "," + z + "),\n"
    cylinder_string += "\t\tradius=" + radius + ",\n"
    cylinder_string += "\t\tlength=" + z_span + ",\n"
    cylinder_string += "\t),\n"
    cylinder_string += "\tmedium=" + medium + ",\n"
    cylinder_string += "\tname='" + name + "'\n"
    cylinder_string += ")\n"
    return cylinder_string, name.replace(" ", "_")


def _addpoly(Lines, i: int, polyslab: int):
    vertices = "[ # NOTE: please specify ]"
    z_min, z_max = "0", "0"
    z_span = "z_min - z_max"
    axis = "2"
    name = "polyslab_" + str(polyslab)
    medium = "td.Medium()"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:10] == "z min":
            z_min = Lines[i][12:].split(";")[0][:-1]
        elif Lines[i][5:10] == "z max":
            z_max = Lines[i][12:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            z_span = Lines[i][13:].split(";")[0][:-1]
            z_min, z_max = "-" + z_span + "/2", z_span + "/2"
        elif Lines[i][5:10] == "index":
            index = Lines[i][12:].split(";")[0][:-1]
            medium = "td.Medium(permittivity=" + index + "**2)"
        elif Lines[i][5:13] == "material":
            medium = Lines[i][15:].split(";")[0][:-1] + " # NOTE: check material library"
        elif Lines[i][5:13] == "vertices":
            vertices = Lines[i][15:].split(";")[0][:-1]
        i += 1
        if i == len(Lines):
            break
    poly_string = name.replace(" ", "_") + " = td.Structure(\n"
    poly_string += "\tgeometry=td.PolySlab(\n"
    poly_string += "\t\tvertices=" + vertices + ",\n"
    poly_string += "\t\taxis=" + axis + ",\n"
    z_min, z_max = _to_um([z_min, z_max])
    poly_string += "\t\tslab_bounds=(" + z_min + "," + z_max + "),\n"
    poly_string += "\t),\n"
    poly_string += "\tmedium=" + medium + ",\n"
    poly_string += "\tname='" + name + "'\n"
    poly_string += ")\n"
    return poly_string, name.replace(" ", "_")


def _adddipole(Lines, i: int, ptdipole: int):
    # https://optics.ansys.com/hc/en-us/articles/360034382854-Plane-wave-and-beam-source-Simulation-object
    x, y, z = "0", "0", "0"
    name = "point_dipole_" + str(ptdipole)
    polarization = "Ez"
    # ---------------- source parameters ----------------
    source_name = "dipole_source_time_" + str(ptdipole)
    wavelength_start = "0.4"
    wavelength_stop = "0.7"
    freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
    fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
    offset = "5.0"
    # ---------------------------------------------------
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:21] == "wavelength start":
            wavelength_start = "(" + Lines[i][23:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:20] == "wavelength stop":
            wavelength_stop = "(" + Lines[i][22:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "frequency":
            freq0 = Lines[i][16:].split(";")[0][:-1] + "*1e12"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "bandwidth":
            fwidth = Lines[i][16:].split(";")[0][:-1] + "*1e12"
        elif Lines[i][5:11] == "offset":
            offset = Lines[i][13:].split(";")[0][:-1] + "*1e-15"
        i += 1
        if i == len(Lines):
            break
    dipole_string = source_name + " = td.GaussianPulse(\n"
    dipole_string += "\tfreq0=" + freq0 + ",\n"
    dipole_string += "\tfwidth=" + fwidth + ",\n"
    dipole_string += "\toffset=" + offset + "\n"
    dipole_string += ")\n"
    dipole_string += name.replace(" ", "_") + " = td.PointDipole(\n"
    x, y, z = _to_um([x, y, z])
    dipole_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    dipole_string += "\tsource_time=" + source_name + ",\n"
    dipole_string += "\tname='" + name + "',\n"
    dipole_string += "\tpolarization='" + polarization + "',\n"
    dipole_string += ")\n"
    return dipole_string, name.replace(" ", "_")


def _addgaussian(Lines, i: int, gauss: int):
    # https://optics.ansys.com/hc/en-us/articles/360034382854-Plane-wave-and-beam-source-Simulation-object
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    name = "gaussian_beam_" + str(gauss)
    direction = "'+'"
    angle_theta = "0"
    angle_phi = "0"
    pol_angle = "0"
    waist_radius = "1.0"
    waist_distance = "0"
    # ---------------- source parameters ----------------
    source_name = "gauss_source_time_" + str(gauss)
    wavelength_start = "0.4"
    wavelength_stop = "0.7"
    freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
    fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
    offset = "5.0"
    # ---------------------------------------------------
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:14] == "direction":
            if Lines[i][16:].split("'")[0] == "backward":
                direction = "'-'"
        elif Lines[i][5:16] == "angle theta":
            angle_theta = Lines[i][18:].split(";")[0][:-1]
        elif Lines[i][5:14] == "angle phi":
            angle_phi = Lines[i][16:].split(";")[0][:-1]
        elif Lines[i][5:22] == "polarization angle":
            pol_angle = Lines[i][24:].split(";")[0][:-1]
        elif Lines[i][5:20] == "waist radius w0":
            waist_radius = Lines[i][22:].split(";")[0][:-1]
        elif Lines[i][5:24] == "distance from waist":
            waist_distance = Lines[i][26:].split(";")[0][:-1]
        elif Lines[i][5:21] == "wavelength start":
            wavelength_start = "(" + Lines[i][23:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:20] == "wavelength stop":
            wavelength_stop = "(" + Lines[i][22:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "frequency":
            freq0 = Lines[i][16:].split(";")[0][:-1] + "e12"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "bandwidth":
            fwidth = Lines[i][16:].split(";")[0][:-1] + "e12"
        elif Lines[i][5:11] == "offset":
            offset = Lines[i][13:].split(";")[0][:-1] + "e-15"
        i += 1
        if i == len(Lines):
            break
    gauss_string = source_name + " = td.GaussianPulse(\n"
    gauss_string += "\tfreq0=" + freq0 + ",\n"
    gauss_string += "\tfwidth=" + fwidth + ",\n"
    gauss_string += "\toffset=" + offset + "\n"
    gauss_string += ")\n"
    gauss_string += name.replace(" ", "_") + " = td.GaussianBeam(\n"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    gauss_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    gauss_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    gauss_string += "\tsource_time=" + source_name + ",\n"
    gauss_string += "\tname='" + name + "',\n"
    gauss_string += "\tdirection=" + direction + ",\n"
    gauss_string += "\tangle_theta=" + angle_theta + ",\n"
    gauss_string += "\tangle_phi=" + angle_phi + ",\n"
    gauss_string += "\tpol_angle=" + pol_angle + ",\n"
    gauss_string += "\twaist_radius=" + waist_radius + "*1e6,\n"
    gauss_string += "\twaist_distance=" + waist_distance + "*1e6,\n"
    gauss_string += ")\n"
    return gauss_string, name.replace(" ", "_")


def _addplane(Lines, i: int, planewv: int):
    # https://optics.ansys.com/hc/en-us/articles/360034382854-Plane-wave-and-beam-source-Simulation-object
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    name = "plane_source_" + str(planewv)
    direction = "'+'"
    angle_theta = "0"
    angle_phi = "0"
    pol_angle = "0"
    # ---------------- source parameters ----------------
    source_name = "plane_source_time_" + str(planewv)
    wavelength_start = "0.4"
    wavelength_stop = "0.7"
    freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
    fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
    offset = "5.0"
    # ---------------------------------------------------
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:14] == "direction":
            if Lines[i][16:].split("'")[0] == "backward":
                direction = "'-'"
        elif Lines[i][5:16] == "angle theta":
            angle_theta = Lines[i][18:].split(";")[0][:-1]
        elif Lines[i][5:14] == "angle phi":
            angle_phi = Lines[i][16:].split(";")[0][:-1]
        elif Lines[i][5:22] == "polarization angle":
            pol_angle = Lines[i][24:].split(";")[0][:-1]
        elif Lines[i][5:21] == "wavelength start":
            wavelength_start = "(" + Lines[i][23:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:20] == "wavelength stop":
            wavelength_stop = "(" + Lines[i][22:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "frequency":
            freq0 = Lines[i][16:].split(";")[0][:-1] + "e12"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "bandwidth":
            fwidth = Lines[i][16:].split(";")[0][:-1] + "e12"
        elif Lines[i][5:11] == "offset":
            offset = Lines[i][13:].split(";")[0][:-1] + "e-15"
        i += 1
        if i == len(Lines):
            break
    plane_string = source_name + " = td.GaussianPulse(\n"
    plane_string += "\tfreq0=" + freq0 + ",\n"
    plane_string += "\tfwidth=" + fwidth + ",\n"
    plane_string += "\toffset=" + offset + "\n"
    plane_string += ")\n"
    plane_string += name.replace(" ", "_") + " = td.PlaneWave(\n"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    plane_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    plane_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    plane_string += "\tsource_time=" + source_name + ",\n"
    plane_string += "\tname='" + name + "',\n"
    plane_string += "\tdirection=" + direction + ",\n"
    plane_string += "\tangle_theta=" + angle_theta + ",\n"
    plane_string += "\tangle_phi=" + angle_phi + ",\n"
    plane_string += "\tpol_angle=" + pol_angle + ",\n"
    plane_string += ")\n"
    return plane_string, name.replace(" ", "_")


def _addtfsf(Lines, i: int, tfsf: int):
    # https://optics.ansys.com/hc/en-us/articles/360034382854-Plane-wave-and-beam-source-Simulation-object
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    name = "tfsf_" + str(tfsf)
    direction = "'+'"
    angle_theta = "0"
    angle_phi = "0"
    pol_angle = "0"
    # ---------------- source parameters ----------------
    source_name = "tfsf_source_time_" + str(tfsf)
    wavelength_start = "0.4"
    wavelength_stop = "0.7"
    freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
    fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
    offset = "5.0"
    # ---------------------------------------------------
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:14] == "direction":
            if Lines[i][16:].split("'")[0] == "backward":
                direction = "'-'"
        elif Lines[i][5:16] == "angle theta":
            angle_theta = Lines[i][18:].split(";")[0][:-1]
        elif Lines[i][5:14] == "angle phi":
            angle_phi = Lines[i][16:].split(";")[0][:-1]
        elif Lines[i][5:22] == "polarization angle":
            pol_angle = Lines[i][24:].split(";")[0][:-1]
        elif Lines[i][5:20] == "waist radius w0":
            waist_radius = Lines[i][22:].split(";")[0][:-1]
        elif Lines[i][5:24] == "distance from waist":
            waist_distance = Lines[i][26:].split(";")[0][:-1]
        elif Lines[i][5:21] == "wavelength start":
            wavelength_start = "(" + Lines[i][23:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:20] == "wavelength stop":
            wavelength_stop = "(" + Lines[i][22:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "frequency":
            freq0 = Lines[i][16:].split(";")[0][:-1] + "e12"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "bandwidth":
            fwidth = Lines[i][16:].split(";")[0][:-1] + "e12"
        elif Lines[i][5:11] == "offset":
            offset = Lines[i][13:].split(";")[0][:-1] + "e-15"
        i += 1
        if i == len(Lines):
            break
    tfsf_string = source_name + " = td.GaussianPulse(\n"
    tfsf_string += "\tfreq0=" + freq0 + ",\n"
    tfsf_string += "\tfwidth=" + fwidth + ",\n"
    tfsf_string += "\toffset=" + offset + "\n"
    tfsf_string += ")\n"
    tfsf_string += name.replace(" ", "_") + " = td.TFSF(\n"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    tfsf_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    tfsf_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    tfsf_string += "\tsource_time=" + source_name + ",\n"
    tfsf_string += "\tname='" + name + "',\n"
    tfsf_string += "\tdirection=" + direction + ",\n"
    tfsf_string += "\tangle_theta=" + angle_theta + ",\n"
    tfsf_string += "\tangle_phi=" + angle_phi + ",\n"
    tfsf_string += "\tpol_angle=" + pol_angle + ",\n"
    tfsf_string += "\twaist_radius=" + waist_radius + "*1e6,\n"
    tfsf_string += "\twaist_distance=" + waist_distance + "*1e6,\n"
    tfsf_string += ")\n"
    return tfsf_string, name.replace(" ", "_")


def _addmode(Lines, i: int, modesrc: int):
    # https://optics.ansys.com/hc/en-us/articles/360034902153-Mode-source-Simulation-object
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    name = "mode_source_" + str(modesrc)
    direction = "'+'"
    num_modes = "1"
    angle_theta = "0"
    mode_spec_name = "mode_spec_" + str(modesrc)
    mode_index = "0"
    # ---------------- source parameters ----------------
    source_name = "mode_source_time_" + str(modesrc)
    wavelength_start = "0.4"
    wavelength_stop = "0.7"
    freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
    fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
    offset = "5.0"
    # ---------------------------------------------------
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:13] == "direction":
            if Lines[i][16:].split("'")[0] == "Backward":
                direction = "'-'"
        elif Lines[i][5:23] == "polarization angle":
            angle_theta = Lines[i][25:].split(";")[0][:-1]
        elif Lines[i][5:19] == "mode selection":
            if Lines[i][22:].split(";")[0][:-1] == "'fundamental TM mode'":
                num_modes, mode_index = "2", "1"
            elif Lines[i][22:].split("'")[0] == "user select":
                num_modes, mode_index = (
                    "PLEASE SPECIFY NUMBER OF MODES",
                    "PLEASE SPECIFY MODE INDEX",
                )
        elif Lines[i][5:21] == "wavelength start":
            wavelength_start = "(" + Lines[i][23:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:20] == "wavelength stop":
            wavelength_stop = "(" + Lines[i][22:].split(";")[0][:-1] + "*1e6)"
            freq0 = "(td.C_0/" + wavelength_start + " + td.C_0/" + wavelength_stop + ")/2"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "frequency":
            freq0 = Lines[i][16:].split(";")[0][:-1] + "e12"
            if wavelength_start == wavelength_stop:
                fwidth = freq0 + "/10"
            else:
                fwidth = "(td.C_0/" + wavelength_start + " - td.C_0/" + wavelength_stop + ")"
        elif Lines[i][5:14] == "bandwidth":
            fwidth = Lines[i][16:].split(";")[0][:-1] + "e12"
        elif Lines[i][5:11] == "offset":
            offset = Lines[i][13:].split(";")[0][:-1] + "e-15"
        i += 1
        if i == len(Lines):
            break
    mode_string = source_name + " = td.GaussianPulse(\n"
    mode_string += "\tfreq0=" + freq0 + ",\n"
    mode_string += "\tfwidth=" + fwidth + ",\n"
    mode_string += "\toffset=" + offset + "\n"
    mode_string += ")\n"
    mode_string += mode_spec_name + " = td.ModeSpec(num_modes=" + num_modes
    mode_string += ",angle_theta=" + angle_theta + ")\n"
    mode_string += name.replace(" ", "_") + " = td.ModeSource(\n"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    mode_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    mode_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    mode_string += "\tsource_time=" + source_name + ",\n"
    mode_string += "\tname='" + name + "',\n"
    mode_string += "\tmode_spec=" + mode_spec_name + ",\n"
    if mode_index != -1:
        mode_string += "\tmode_index=" + mode_index + ",\n"
    mode_string += "\tdirection=" + direction + ",\n"
    mode_string += ")\n"
    return mode_string, name.replace(" ", "_")


def _get_BCs(dim_bc_file: str):
    dim_bc, dim_sym = "td.PML(num_layers=", "0"
    if dim_bc_file == "Periodic":
        dim_bc = "td.Periodic()"
    elif dim_bc_file == "Metal":
        dim_bc = "td.PECBoundary()"
    elif dim_bc_file == "Symmetric":
        dim_sym = "1"
    elif dim_bc_file == "Anti-Symmetric":
        dim_sym = "-1"
    elif dim_bc_file == "PMC":
        dim_bc = "td.PMCBoundary()"
    elif dim_bc_file == "Bloch":
        dim_bc = "td.BlochBoundary(bloch_vec=1)"
    return dim_bc, dim_sym


def _addfdtd(Lines, i: int, v_dict: dict) -> str:
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    x_min, x_max = "0", "0"
    y_min, y_max = "0", "0"
    z_min, z_max = "0", "0"
    background_index = "1"
    runtime = " # NOTE: please specify"
    mesh_accuracy = 1
    min_steps_per_wvl = 6
    num_layers = "12"
    x_min_bc = "td.PML(num_layers="
    x_max_bc = x_min_bc
    y_min_bc, y_max_bc = x_min_bc, x_min_bc
    z_min_bc, z_max_bc = x_min_bc, x_min_bc
    x_sym, y_sym, z_sym = "0", "0", "0"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:10] == "x_min":
            x_min = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "x_max":
            x_max = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "y_min":
            y_min = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "y_max":
            y_max = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "z_min":
            z_min = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:10] == "z_max":
            z_max = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:21] == "background index":
            background_index = Lines[i][23:].split(";")[0][:-1]
        elif Lines[i][5:18] == "mesh accuracy":
            mesh_accuracy = v_dict[Lines[i][20:].split(";")[0][:-1]]
            min_steps_per_wvl = str(int(2 + 4 * mesh_accuracy))
        elif Lines[i][5:13] == "x min bc":  # min
            x_min_bc_file = Lines[i][16:].split("'")[0]
            x_min_bc, x_sym = _get_BCs(x_min_bc_file)
        elif Lines[i][5:13] == "y min bc":
            y_min_bc_file = Lines[i][16:].split("'")[0]
            y_min_bc, y_sym = _get_BCs(y_min_bc_file)
        elif Lines[i][5:13] == "z min bc":
            z_min_bc_file = Lines[i][16:].split("'")[0]
            z_min_bc, z_sym = _get_BCs(z_min_bc_file)
        elif Lines[i][5:13] == "x max bc":  # max
            x_max_bc_file = Lines[i][16:].split("'")[0]
            x_max_bc, x_sym = _get_BCs(x_max_bc_file)
        elif Lines[i][5:13] == "y max bc":
            y_max_bc_file = Lines[i][16:].split("'")[0]
            y_max_bc, y_sym = _get_BCs(y_max_bc_file)
        elif Lines[i][5:13] == "z max bc":
            z_max_bc_file = Lines[i][16:].split("'")[0]
            z_max_bc, z_sym = _get_BCs(z_max_bc_file)
        elif Lines[i][5:15] == "pml layers":
            num_layers = Lines[i][17:].split(";")[0][:-1]
        elif Lines[i][5:20] == "simulation time":
            runtime = Lines[i][22:].split(";")[0][:-1]
        i += 1
        if i == len(Lines):
            break
    if xl in v_dict:
        if v_dict[xl] == 0:
            x_min_bc, x_max_bc = "td.Periodic()", "td.Periodic()"
    elif xl == "0":
        x_min_bc, x_max_bc = "td.Periodic()", "td.Periodic()"
    if yl in v_dict:
        if v_dict[yl] == 0:
            y_min_bc, y_max_bc = "td.Periodic()", "td.Periodic()"
    elif yl == "0":
        y_min_bc, y_max_bc = "td.Periodic()", "td.Periodic()"
    if zl in v_dict:
        if v_dict[zl] == 0:
            z_min_bc, z_max_bc = "td.Periodic()", "td.Periodic()"
    elif zl == "0":
        z_min_bc, z_max_bc = "td.Periodic()", "td.Periodic()"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    sim_string = "sim = td.Simulation(\n"
    sim_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    sim_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    sim_string += "\trun_time=" + runtime + ",\n"
    sim_string += "\tmedium=td.Medium(permittivity=" + background_index + "**2),\n"
    sim_string += "\tsymmetry=(" + x_sym + "," + y_sym + "," + z_sym + "),\n"
    sim_string += "\tboundary_spec=td.BoundarySpec(\n"
    sim_string += "\t\tx = td.Boundary(minus=" + x_min_bc
    if x_min_bc[-1] == "=":  # it's a pml boundary
        sim_string += num_layers + ")"
    sim_string += ", plus=" + x_max_bc
    if x_max_bc[-1] == "=":  # ti's a pml boundary
        sim_string += num_layers + ")"
    sim_string += "),\n"
    sim_string += "\t\ty = td.Boundary(minus=" + y_min_bc
    if y_min_bc[-1] == "=":  # it's a pml boundary
        sim_string += num_layers + ")"
    sim_string += ", plus=" + y_max_bc
    if y_max_bc[-1] == "=":  # ti's a pml boundary
        sim_string += num_layers + ")"
    sim_string += "),\n"
    sim_string += "\t\tz = td.Boundary(minus=" + z_min_bc
    if z_min_bc[-1] == "=":  # it's a pml boundary
        sim_string += num_layers + ")"
    sim_string += ", plus=" + z_max_bc
    if z_max_bc[-1] == "=":  # ti's a pml boundary
        sim_string += num_layers + ")"
    sim_string += "),\n\t),\n"
    sim_string += "\t"
    sim_string += "grid_spec="
    gridspec_string = "td.GridSpec.auto(min_steps_per_wvl=" + min_steps_per_wvl + ","
    gridspec_string += " wavelength=1.0)"
    sim_string += gridspec_string
    sim_string += ",\n)\n"
    return sim_string, gridspec_string


def _addmesh(Lines, i: int, overstrct: int):
    x, y, z = "0", "0", "0"
    xl, yl, zl = "td.inf", "td.inf", "td.inf"
    dx, dy, dz = "None", "None", "None"
    name = "refine_box_" + str(overstrct)
    override_string = " = td.MeshOverrideStructure(\n"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:7] == "dx":
            dx = Lines[i][9:].split(";")[0][:-1]
        elif Lines[i][5:7] == "dy":
            dy = Lines[i][9:].split(";")[0][:-1]
        elif Lines[i][5:7] == "dz":
            dz = Lines[i][9:].split(";")[0][:-1]
        i += 1
        if i == len(Lines):
            break
    x, y, z, xl, yl, zl, dx, dy, dz = _to_um([x, y, z, xl, yl, zl, dx, dy, dz])
    override_string += "\tgeometry=td.Box(center=(" + x + "," + y + "," + z + "),"
    override_string += "size=(" + xl + "," + yl + "," + zl + ")),\n"
    override_string += "\tdl=(" + dx + "," + dy + "," + dz + "),\n)\n"
    return override_string, name.replace(" ", "_")


def _addindex(Lines, i: int, indexmon: int):
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    name = "permittivity_monitor_" + str(indexmon)
    freqs = " # NOTE: please specify"
    wave_center, wave_span = "0.5", "0.1"
    index_monitor_string = " = td.PermittivityMonitor(\n"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:22] == "wavelength center":
            wave_center = Lines[i][24:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:20] == "wavelength span":
            wave_span = Lines[i][22:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:21] == "frequency points":
            num_freqs = Lines[i][23:].split(";")[0][:-1]
            freqs = (
                "np.linspace("
                + wave_center
                + "-"
                + wave_span
                + "/2,"
                + wave_center
                + "+"
                + wave_span
                + "/2,"
                + num_freqs
                + ")"
            )
        i += 1
        if i == len(Lines):
            break
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    index_monitor_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    index_monitor_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    index_monitor_string += "\tname='" + name + "',\n"
    index_monitor_string += "\tinterval_space=(1,1,1),\n"
    index_monitor_string += "\tfreqs=" + freqs + ",\n"
    index_monitor_string += ")\n"
    return index_monitor_string, name.replace(" ", "_")


def _addpower(Lines, i: int, fluxmon: int):
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    name = "flux_monitor_" + str(fluxmon)
    freqs = " # NOTE: please specify"
    wave_center, wave_span = "0.5", "0.1"
    flux_monitor_string = " = td.FluxMonitor(\n"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:22] == "wavelength center":
            wave_center = Lines[i][24:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:20] == "wavelength span":
            wave_span = Lines[i][22:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:21] == "frequency points":
            num_freqs = Lines[i][23:].split(";")[0][:-1]
            freqs = (
                "np.linspace("
                + wave_center
                + "-"
                + wave_span
                + "/2,"
                + wave_center
                + "+"
                + wave_span
                + "/2,"
                + num_freqs
                + ")"
            )
        i += 1
        if i == len(Lines):
            break
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    flux_monitor_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    flux_monitor_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    flux_monitor_string += "\tname='" + name + "',\n"
    flux_monitor_string += "\tfreqs=" + freqs + ",\n"
    flux_monitor_string += ")\n"
    return flux_monitor_string, name.replace(" ", "_")


def _addmovie(Lines, i: int, fieldTime: int):
    # https://optics.ansys.com/hc/en-us/articles/360034924633-addefieldmonitor
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    x_min, x_max = "0", "0"
    y_min, y_max = "0", "0"
    z_min, z_max = "0", "0"
    name = "field_time_monitor_" + str(fieldTime)
    field_monitor_string = " = td.FieldTimeMonitor(\n"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:10] == "x_min":
            x_min = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "x_max":
            x_max = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "y_min":
            y_min = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "y_max":
            y_max = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "z_min":
            z_min = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:10] == "z_max":
            z_max = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        i += 1
        if i == len(Lines):
            break
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    field_monitor_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    field_monitor_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    field_monitor_string += "\tname='" + name + "',\n"
    field_monitor_string += "\tinterval=1,\n"
    field_monitor_string += "\tstart=0.0,\n"
    field_monitor_string += "\tstop=None,\n"
    field_monitor_string += ")\n"
    return field_monitor_string, name.replace(" ", "_")


def _addefieldmonitor(Lines, i: int, fieldmon: int):
    # https://optics.ansys.com/hc/en-us/articles/360034924633-addefieldmonitor
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    x_min, x_max = "0", "0"
    y_min, y_max = "0", "0"
    z_min, z_max = "0", "0"
    name = "field_monitor_" + str(fieldmon)
    freqs = " # NOTE: please specify"
    wave_center, wave_span = "0.5", "0.1"
    field_monitor_string = " = td.FieldMonitor(\n"
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:10] == "x_min":
            x_min = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "x_max":
            x_max = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "y_min":
            y_min = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "y_max":
            y_max = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "z_min":
            z_min = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:10] == "z_max":
            z_max = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:22] == "wavelength center":
            wave_center = Lines[i][24:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:20] == "wavelength span":
            wave_span = Lines[i][22:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:21] == "frequency points":
            num_freqs = Lines[i][23:].split(";")[0][:-1]
            freqs = (
                "np.linspace("
                + wave_center
                + "-"
                + wave_span
                + "/2,"
                + wave_center
                + "+"
                + wave_span
                + "/2,"
                + num_freqs
                + ")"
            )
        # add monitor type if I ever find mapping between the numbers?
        i += 1
        if i == len(Lines):
            break
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    field_monitor_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    field_monitor_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    field_monitor_string += "\tname='" + name + "',\n"
    field_monitor_string += "\tfreqs=" + freqs + ",\n"
    field_monitor_string += ")\n"
    return field_monitor_string, name.replace(" ", "_")


def _addmodeexpansion(Lines, i: int, modemon: int):
    # https://optics.ansys.com/hc/en-us/articles/360034902413
    x, y, z = "0", "0", "0"
    xl, yl, zl = "0", "0", "0"
    x_min, x_max = "0", "0"
    y_min, y_max = "0", "0"
    z_min, z_max = "0", "0"
    name = "mode_monitor_" + str(modemon)
    num_modes = "1"
    mode = "0"
    freqs = " # NOTE: please specify"
    wave_center, wave_span = "0.5", "0.1"
    mode_spec_name = "mode_spec_" + str(modemon)
    mode_monitor_string = mode_spec_name + " = td.ModeSpec(num_modes="
    while Lines[i][0] == "\n" or Lines[i][:5] == "set('" or Lines[i][0] == "#":
        if Lines[i][5:9] == "name":
            name = Lines[i][12:].split("'")[0]
        elif Lines[i][5:7] == "x'":
            x = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "y'":
            y = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:7] == "z'":
            z = Lines[i][8:].split(";")[0][:-1]
        elif Lines[i][5:10] == "x_min":
            x_min = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "x_max":
            x_max = Lines[i][12:].split(";")[0][:-1]
            xl = x_max + "-" + x_min
        elif Lines[i][5:10] == "y_min":
            y_min = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "y_max":
            y_max = Lines[i][12:].split(";")[0][:-1]
            yl = y_max + "-" + y_min
        elif Lines[i][5:10] == "z_min":
            z_min = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:10] == "z_max":
            z_max = Lines[i][12:].split(";")[0][:-1]
            zl = z_max + "-" + z_min
        elif Lines[i][5:11] == "x span":
            xl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "y span":
            yl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:11] == "z span":
            zl = Lines[i][13:].split(";")[0][:-1]
        elif Lines[i][5:19] == "mode selection":
            mode = Lines[i][22:].split(";")[0][:-1]
            if mode == "'fundamental TM mode'":
                mode = "1"
                num_modes = "2"
            else:
                mode = "0"
        elif Lines[i][5:22] == "wavelength center":
            wave_center = Lines[i][24:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:20] == "wavelength span":
            wave_span = Lines[i][22:].split(";")[0][:-1] + "*1e6"
        elif Lines[i][5:21] == "frequency points":
            num_freqs = Lines[i][23:].split(";")[0][:-1]
            freqs = (
                "np.linspace("
                + wave_center
                + "-"
                + wave_span
                + "/2,"
                + wave_center
                + "+"
                + wave_span
                + "/2,"
                + num_freqs
                + ")"
            )
        i += 1
        if i == len(Lines):
            break
    mode_monitor_string += num_modes + ") # NOTE: check if solving for correct mode!\n"
    mode_monitor_string += name.replace(" ", "_") + " = td.ModeMonitor(\n"
    x, y, z, xl, yl, zl = _to_um([x, y, z, xl, yl, zl])
    mode_monitor_string += "\tcenter=(" + x + "," + y + "," + z + "),\n"
    mode_monitor_string += "\tsize=(" + xl + "," + yl + "," + zl + "),\n"
    mode_monitor_string += "\tfreqs=" + freqs + ",\n"
    mode_monitor_string += "\tmode_spec=" + mode_spec_name + ",\n"
    mode_monitor_string += "\tname='" + name + "',\n"
    mode_monitor_string += ")\n"
    return mode_monitor_string, name.replace(" ", "_")


def lsf_reader(filename: str) -> None:
    tidy3d_file = "import numpy as np\n"
    tidy3d_file += "import matplotlib.pyplot as plt\n"
    tidy3d_file += "import tidy3d as td\n"
    tidy3d_file += "import tidy3d.web as web\n"
    tidy3d_file += "\n"

    structures = "["
    rect, sph, cyl, polyslab = 0, 0, 0, 0
    sources, modesrc, planewv, ptdipole, gauss, tfsf = "[", 0, 0, 0, 0, 0
    monitors = "["
    modemon, fluxmon, fieldmon, fieldTime, indexmon = 0, 0, 0, 0, 0
    gridspec_string = "grid_spec=td.GridSpec.auto()"
    override_structures, overstrct = "[", 0

    with open(filename) as file_lsf:
        Lines = file_lsf.readlines()
    variable_dict = {}
    for i, line in enumerate(Lines):
        Lines[i] = _remove_spaces(line)
    for i, line in enumerate(Lines):
        line = Lines[i]
        if line[0] == "#":
            tidy3d_file += line
        elif line[0] == "\n":
            tidy3d_file += line
        elif _is_declaration(line):
            _add_to_variable_dict(line, variable_dict)
            tidy3d_file += line.replace(";", " ")
        elif line[:8] == "addrect;":
            addrect_string, rect_name = _addrect(Lines, i + 1, rect)
            tidy3d_file += addrect_string
            structures += rect_name + ", "
            rect += 1
        elif line[:10] == "addsphere;":
            addsphere_string, sphere_name = _addsphere(Lines, i + 1, sph)
            tidy3d_file += addsphere_string
            structures += sphere_name + ", "
            sph += 1
        elif line[:10] == "addcircle;":
            addcircle_string, cylinder_name = _addcircle(Lines, i + 1, cyl)
            tidy3d_file += addcircle_string
            structures += cylinder_name + ", "
            cyl += 1
        elif line[:8] == "addpoly;":
            addpoly_string, poly_name = _addpoly(Lines, i + 1, polyslab)
            tidy3d_file += addpoly_string
            structures += poly_name + ", "
            polyslab += 1
        elif line[:10] == "adddipole;":
            adddipole_string, dipole_name = _adddipole(Lines, i + 1, ptdipole)
            tidy3d_file += adddipole_string
            sources += dipole_name + ", "
            ptdipole += 1
        elif line[:12] == "addgaussian;":
            addgaussian_string, gaussian_name = _addgaussian(Lines, i + 1, gauss)
            tidy3d_file += addgaussian_string
            sources += gaussian_name + ", "
            gauss += 1
        elif line[:9] == "addplane;":
            addplane_string, plane_name = _addplane(Lines, i + 1, planewv)
            tidy3d_file += addplane_string
            sources += plane_name + ", "
            planewv += 1
        elif line[:8] == "addtfsf;":
            addtfsf_string, tfsf_name = _addtfsf(Lines, i + 1, tfsf)
            tidy3d_file += addtfsf_string
            sources += tfsf_name + ", "
            tfsf += 1
        elif line[:8] == "addmode;":
            addmode_string, mode_src_name = _addmode(Lines, i + 1, modesrc)
            tidy3d_file += addmode_string
            sources += mode_src_name + ", "
            modesrc += 1
        elif line[:8] == "addfdtd;":
            addfdtd_string, gridspec_string = _addfdtd(Lines, i + 1, variable_dict)
            tidy3d_file += addfdtd_string
        elif line[:8] == "addmesh;":
            override_string, refine_box_name = _addmesh(Lines, i + 1, overstrct)
            override_structures += refine_box_name + ", "
            tidy3d_file += refine_box_name + override_string
            overstrct += 1
        elif (
            line[:9] == "addindex;"
            or line[:18] == "addeffectiveindex;"
            or line[:12] == "addemeindex;"
        ):
            addindex_string, index_name = _addindex(Lines, i + 1, indexmon)
            monitors += index_name + ", "
            tidy3d_file += index_name + addindex_string
            indexmon += 1
        elif line[:9] == "addpower;":
            addpower_string, power_name = _addpower(Lines, i + 1, fluxmon)
            monitors += power_name + ", "
            tidy3d_file += power_name + addpower_string
            fluxmon += 1
        elif (
            line[:9] == "addmovie;"
            or line[:22] == "addemfieldtimemonitor;"
            or line[:8] == "addtime;"
        ):
            addmovie_string, movie_name = _addmovie(Lines, i + 1, fieldTime)
            monitors += movie_name + ", "
            tidy3d_file += movie_name + addmovie_string
            fieldTime += 1
        elif (
            line[:17] == "addefieldmonitor;"
            or line[:11] == "addprofile;"
            or line[:18] == "addemfieldmonitor;"
            or line[:14] == "addemeprofile;"
        ):
            addefieldmonitor_string, monitor_name = _addefieldmonitor(Lines, i + 1, fieldmon)
            monitors += monitor_name + ", "
            tidy3d_file += monitor_name + addefieldmonitor_string
            fieldmon += 1
        elif line[:17] == "addmodeexpansion;":
            addmodeexpansion_string, modemonitor_name = _addmodeexpansion(Lines, i + 1, modemon)
            # not adding name first as we needed to create mode_spec
            tidy3d_file += addmodeexpansion_string
            monitors += modemonitor_name + ", "
            modemon += 1
        elif line[:5] == "set('":
            pass
        elif line[:14] == "switchtolayout":
            pass
        elif line[:9] == "selectall":
            pass
        elif line[:6] == "delete":
            pass
        else:
            tidy3d_file += "# " + line[:-1] + " # NOTE: does not yet parse to Tidy3D\n"
    if len(structures) > 1:
        structures = structures[:-2] + "]"
    else:
        structures = "[]"
    if len(sources) > 1:
        sources = sources[:-2] + "]"
    else:
        sources = "[]"
    if len(monitors) > 1:
        monitors = monitors[:-2] + "]"
    else:
        monitors = "[]"
    if len(override_structures) > 1:
        override_structures = override_structures[:-2] + "]"
    else:
        override_structures = "[]"
    tidy3d_file += "\nsim = sim.copy(update=dict(structures=" + structures + ","
    tidy3d_file += " # NOTE: Check order of structures for potential overlap issues\n"
    tidy3d_file += "\tsources=" + sources + ",\n"
    tidy3d_file += "\tmonitors=" + monitors + ",\n"
    tidy3d_file += (
        "\tgrid_spec="
        + gridspec_string[:-1]
        + ", override_structures="
        + override_structures
        + ")\n\t)\n"
    )
    tidy3d_file += ")"
    return tidy3d_file


def converter_arg(lsf_file, new_file):
    if Path(f"{lsf_file}").suffix != ".lsf":
        raise Exception(f"{lsf_file} must be an .lsf file.")
    if Path(f"{new_file}").suffix != ".py":
        raise Exception(f"{new_file} must be a .py file.")
    file_string = lsf_reader(lsf_file)
    num_specs = file_string.count("NOTE: please specify")
    with open(new_file, "w") as f:
        f.write(file_string)
    print("Created new Tidy3D file:", new_file)
    print(
        f"\nNOTE: This is an experimental feature. Not all scripting commands are supported. Some default variables are specified if not given in the .lsf file, and {num_specs} variables still need to be set in the script in order to run. Check the comments in the output file for hints."
    )
    return



"""Objects that define how data is recorded from simulation."""
from abc import ABC
from typing import Union

from ..types import ArrayFloat1D
from ..base_sim.monitor import AbstractMonitor


BYTES_REAL = 4


class HeatMonitor(AbstractMonitor, ABC):
    """Abstract base class for heat monitors."""


class TemperatureMonitor(HeatMonitor):
    """Temperature monitor."""

    def storage_size(self, num_cells: int, tmesh: ArrayFloat1D) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 real number per grid cell, per time step, per field
        num_steps = self.num_steps(tmesh)
        return BYTES_REAL * num_steps * num_cells * len(self.fields)


# types of monitors that are accepted by heat simulation
HeatMonitorType = Union[TemperatureMonitor]



"""Tests space time modulation."""
import numpy as np
import pytest
from math import isclose
import pydantic.v1 as pydantic
import tidy3d as td
from tidy3d.exceptions import ValidationError

np.random.seed(4)

# space
NX, NY, NZ = 10, 9, 8
X = np.linspace(-1, 1, NX)
Y = np.linspace(-1, 1, NY)
Z = np.linspace(-1, 1, NZ)
COORDS = dict(x=X, y=Y, z=Z)
ARRAY_CMP = td.SpatialDataArray(np.random.random((NX, NY, NZ)) + 0.1j, coords=COORDS)
ARRAY = td.SpatialDataArray(np.random.random((NX, NY, NZ)), coords=COORDS)

SP_UNIFORM = td.SpaceModulation()

# time
FREQ_MODULATE = 1e12
AMP_TIME = 1.1
PHASE_TIME = 0
CW = td.ContinuousWaveTimeModulation(freq0=FREQ_MODULATE, amplitude=AMP_TIME, phase=PHASE_TIME)

# combined
ST = td.SpaceTimeModulation(
    time_modulation=CW,
)

# medium modulation spec
MODULATION_SPEC = td.ModulationSpec()


def test_time_modulation():
    """time modulation: only supporting CW for now."""
    assert isclose(np.real(CW.amp_time(1 / FREQ_MODULATE)), AMP_TIME)
    assert isclose(CW.max_modulation, AMP_TIME)

    cw = CW.updated_copy(phase=np.pi / 4, amplitude=10)
    assert isclose(np.real(cw.amp_time(1 / FREQ_MODULATE)), np.sqrt(2) / 2 * 10)
    assert isclose(cw.max_modulation, 10)


def test_space_modulation():
    """uniform or custom space modulation"""
    # uniform in both amplitude and phase
    assert isclose(SP_UNIFORM.max_modulation, 1)

    # uniform in phase, but custom in amplitude
    with pytest.raises(pydantic.ValidationError):
        sp = SP_UNIFORM.updated_copy(amplitude=ARRAY_CMP)

    sp = SP_UNIFORM.updated_copy(amplitude=ARRAY)
    assert isclose(sp.max_modulation, np.max(ARRAY))

    # uniform in amplitude, but custom in phase
    with pytest.raises(pydantic.ValidationError):
        sp = SP_UNIFORM.updated_copy(phase=ARRAY_CMP)
    sp = SP_UNIFORM.updated_copy(phase=ARRAY)
    assert isclose(sp.max_modulation, 1)

    # custom in both
    with pytest.raises(pydantic.ValidationError):
        sp = SP_UNIFORM.updated_copy(phase=ARRAY_CMP, amplitude=ARRAY_CMP)
    sp = SP_UNIFORM.updated_copy(phase=ARRAY, amplitude=ARRAY)


def test_space_time_modulation():
    # cw modulation, uniform in space
    assert isclose(ST.max_modulation, AMP_TIME)
    assert not ST.negligible_modulation

    # cw modulation, but 0 amplitude
    st = ST.updated_copy(time_modulation=CW.updated_copy(amplitude=0))
    assert st.negligible_modulation

    st = ST.updated_copy(space_modulation=td.SpaceModulation(amplitude=0))
    assert st.negligible_modulation

    # cw modulation, nonuniform in space
    st = ST.updated_copy(space_modulation=td.SpaceModulation(amplitude=ARRAY, phase=ARRAY))
    assert not st.negligible_modulation
    assert isclose(st.max_modulation, AMP_TIME * np.max(ARRAY))


def test_modulated_medium():
    """time modulated medium"""
    # unmodulated
    medium = td.Medium()
    assert medium.modulation_spec is None
    assert medium.time_modulated == False

    assert MODULATION_SPEC.applied_modulation == False
    medium = medium.updated_copy(modulation_spec=MODULATION_SPEC)
    assert medium.time_modulated == False

    # permittivity modulated
    modulation_spec = MODULATION_SPEC.updated_copy(permittivity=ST)
    # modulated permitivity <= 0
    with pytest.raises(pydantic.ValidationError):
        medium = td.Medium(modulation_spec=modulation_spec)
    medium = td.Medium(permittivity=2, modulation_spec=modulation_spec)
    assert isclose(medium.n_cfl, np.sqrt(2 - AMP_TIME))

    # conductivity modulated
    modulation_spec = MODULATION_SPEC.updated_copy(conductivity=ST)
    # modulated conductivity <= 0
    with pytest.raises(pydantic.ValidationError):
        medium = td.Medium(modulation_spec=modulation_spec)
    medium_sometimes_active = td.Medium(modulation_spec=modulation_spec, allow_gain=True)
    medium = td.Medium(conductivity=2, modulation_spec=modulation_spec)

    # both modulated, but different time modulation: error
    st_freq2 = ST.updated_copy(
        time_modulation=td.ContinuousWaveTimeModulation(freq0=2e12, amplitude=2)
    )
    with pytest.raises(pydantic.ValidationError):
        modulation_spec = MODULATION_SPEC.updated_copy(permittivity=ST, conductivity=st_freq2)
    # both modulated, but different space modulation: fine
    st_space2 = ST.updated_copy(space_modulation=td.SpaceModulation(amplitude=0.1))
    modulation_spec = MODULATION_SPEC.updated_copy(permittivity=ST, conductivity=st_space2)
    medium = td.Medium(
        permittivity=3,
        conductivity=1,
        modulation_spec=modulation_spec,
    )


def test_unsupported_modulated_medium_types():
    """Unsupported types of time modulated medium"""
    modulation_spec = MODULATION_SPEC.updated_copy(permittivity=ST)

    # PEC cannot be modulated
    with pytest.raises(pydantic.ValidationError):
        mat = td.PECMedium(modulation_spec=modulation_spec)

    # For Anisotropic medium, one should modulate the components, not the whole medium
    with pytest.raises(pydantic.ValidationError):
        mat = td.AnisotropicMedium(
            xx=td.Medium(), yy=td.Medium(), zz=td.Medium(), modulation_spec=modulation_spec
        )

    # Modulation to fully Anisotropic medium unsupported
    with pytest.raises(pydantic.ValidationError):
        mat = td.FullyAnisotropicMedium(modulation_spec=modulation_spec)

    # 2D material
    with pytest.raises(pydantic.ValidationError):
        drude_medium = td.Drude(eps_inf=2.0, coeffs=[(1, 2), (3, 4)])
        medium2d = td.Medium2D(ss=drude_medium, tt=drude_medium, modulation_spec=modulation_spec)

    # together with nonlinear_spec
    with pytest.raises(pydantic.ValidationError):
        mat = td.Medium(
            permittivity=2,
            nonlinear_spec=td.NonlinearSusceptibility(chi3=1),
            modulation_spec=modulation_spec,
        )


def test_supported_modulated_medium_types():
    """Supported types of time modulated medium"""
    modulation_spec = MODULATION_SPEC.updated_copy(permittivity=ST)
    modulation_both_spec = modulation_spec.updated_copy(conductivity=ST)

    # Dispersive
    mat_p = td.PoleResidue(
        eps_inf=2.0, poles=[((-1 + 2j), (3 + 4j))], modulation_spec=modulation_spec
    )
    assert mat_p.time_modulated
    assert isclose(mat_p.n_cfl, np.sqrt(2 - AMP_TIME))
    # too much modulation resulting in eps_inf < 0
    with pytest.raises(pydantic.ValidationError):
        mat = mat_p.updated_copy(eps_inf=1.0)
    # conductivity modulation
    with pytest.raises(pydantic.ValidationError):
        mat = mat_p.updated_copy(modulation_spec=modulation_both_spec)
    mat = mat_p.updated_copy(modulation_spec=modulation_both_spec, allow_gain=True)

    # custom
    permittivity = td.SpatialDataArray(np.ones((1, 1, 1)) * 2, coords=dict(x=[1], y=[1], z=[1]))
    mat_c = td.CustomMedium(permittivity=permittivity, modulation_spec=modulation_spec)
    assert mat_c.time_modulated
    assert isclose(mat_c.n_cfl, np.sqrt(2 - AMP_TIME))
    # too much modulation resulting in eps_inf < 0
    with pytest.raises(pydantic.ValidationError):
        mat = mat_c.updated_copy(permittivity=permittivity * 0.5)
    # conductivity modulation
    with pytest.raises(pydantic.ValidationError):
        mat = mat_c.updated_copy(modulation_spec=modulation_both_spec)
    mat = mat_c.updated_copy(modulation_spec=modulation_both_spec, allow_gain=True)

    # anisotropic medium component
    mat = td.AnisotropicMedium(xx=td.Medium(), yy=mat_p, zz=td.Medium())
    assert mat.time_modulated
    assert isclose(mat.n_cfl, np.sqrt(2 - AMP_TIME))

    # custom anistropic medium component
    mat_uc = td.CustomMedium(permittivity=permittivity)
    mat = td.CustomAnisotropicMedium(xx=mat_uc, yy=mat_c, zz=mat_uc)
    assert mat.time_modulated
    assert isclose(mat.n_cfl, np.sqrt(2 - AMP_TIME))



"""File compression utilities"""

import gzip
import os
import shutil
import tempfile

import h5py

from ..core.constants import JSON_TAG


def compress_file_to_gzip(input_file, output_gz_file):
    """
    Compresses a file using gzip.

    Args:
        input_file (str): The path of the input file.
        output_gz_file (str): The path of the output gzip file.
    """
    with open(input_file, "rb") as file_in:
        with gzip.open(output_gz_file, "wb") as file_out:
            shutil.copyfileobj(file_in, file_out)


def extract_gzip_file(input_gz_file, output_file):
    """
    Extract a gzip file.

    Args:
        input_gz_file (str): The path of the gzip input file.
        output_file (str): The path of the output file.
    """
    with gzip.open(input_gz_file, "rb") as file_in:
        with open(output_file, "wb") as file_out:
            shutil.copyfileobj(file_in, file_out)


def read_simulation_from_hdf5_gz(file_name: str) -> str:
    """read simulation str from hdf5.gz"""

    hdf5_file, hdf5_file_path = tempfile.mkstemp(".hdf5")
    os.close(hdf5_file_path)
    try:
        extract_gzip_file(file_name, hdf5_file_path)
        json_str = read_simulation_from_hdf5(file_name)
    finally:
        os.unlink(hdf5_file_path)
    return json_str


def read_simulation_from_hdf5(file_name: str) -> str:
    """read simulation str from hdf5"""

    with h5py.File(file_name, "r") as f_handle:
        json_string = f_handle[JSON_TAG][()]
        return json_string


def read_simulation_from_json(file_name: str) -> str:
    """read simulation str from json"""
    with open(file_name) as json_file:
        json_data = json_file.read()
    return json_data



"""Defines jax-compatible datasets."""
import pydantic.v1 as pd

from jax.tree_util import register_pytree_node_class

from .....components.data.dataset import PermittivityDataset
from .data_array import JaxDataArray
from ..base import JaxObject


@register_pytree_node_class
class JaxPermittivityDataset(PermittivityDataset, JaxObject):
    """A :class:`.PermittivityDataset` registered with jax."""

    eps_xx: JaxDataArray = pd.Field(
        ...,
        title="Epsilon xx",
        description="Spatial distribution of the xx-component of the relative permittivity.",
        jax_field=True,
    )
    eps_yy: JaxDataArray = pd.Field(
        ...,
        title="Epsilon yy",
        description="Spatial distribution of the yy-component of the relative permittivity.",
        jax_field=True,
    )
    eps_zz: JaxDataArray = pd.Field(
        ...,
        title="Epsilon zz",
        description="Spatial distribution of the zz-component of the relative permittivity.",
        jax_field=True,
    )



"""Test converting .lsf files to Tidy3D python files."""

import pytest
import os
from tidy3d.web.cli.converter import converter_arg


@pytest.mark.parametrize("lsf_file", ("tests/data/example.lsf", "tests/data/monitors.lsf"))
def test_tidy3d_converter(lsf_file, tmp_path):
    """Generates Tidy3D python files from example lsf files in tests/data"""

    new_file_path = str(tmp_path / "test.py")
    converter_arg(lsf_file, new_file_path)
    assert os.path.exists(new_file_path)



"""Tests the scene and its validators."""
import pytest
import pydantic.v1 as pd
import matplotlib.pyplot as plt

import numpy as np
import tidy3d as td
from tidy3d.components.scene import MAX_NUM_MEDIUMS, MAX_GEOMETRY_COUNT
from ..utils import assert_log_level, log_capture, SIM_FULL

SCENE = td.Scene()

SCENE_FULL = SIM_FULL.scene


def test_scene_init():
    """make sure a scene can be initialized"""

    scene = td.Scene(
        structures=[
            td.Structure(
                geometry=td.Box(size=(1, 1, 1), center=(-1, 0, 0)),
                medium=td.Medium(permittivity=2.0),
            ),
            td.Structure(
                geometry=td.Box(size=(1, 1, 1), center=(0, 0, 0)),
                medium=td.Medium(permittivity=1.0, conductivity=3.0),
            ),
            td.Structure(
                geometry=td.Sphere(radius=1.4, center=(1.0, 0.0, 1.0)), medium=td.Medium()
            ),
            td.Structure(
                geometry=td.Cylinder(radius=1.4, length=2.0, center=(1.0, 0.0, -1.0), axis=1),
                medium=td.Medium(),
            ),
        ],
        medium=td.Medium(permittivity=3.0),
    )

    _ = scene.mediums
    _ = scene.medium_map
    _ = scene.background_structure


def test_validate_components_none():

    assert SCENE._validate_num_mediums(val=None) is None


def test_plot_eps():
    ax = SCENE_FULL.plot_eps(x=0)
    SCENE_FULL._add_cbar_eps(eps_min=1, eps_max=2, ax=ax)
    plt.close()


def test_plot_eps_bounds():
    _ = SCENE_FULL.plot_eps(x=0, hlim=[-0.45, 0.45])
    plt.close()
    _ = SCENE_FULL.plot_eps(x=0, vlim=[-0.45, 0.45])
    plt.close()
    _ = SCENE_FULL.plot_eps(x=0, hlim=[-0.45, 0.45], vlim=[-0.45, 0.45])
    plt.close()


def test_plot():
    SCENE_FULL.plot(x=0)
    plt.close()


def test_plot_1d_scene():
    s = td.Scene(structures=[td.Structure(geometry=td.Box(size=(0, 0, 1)), medium=td.Medium())])
    _ = s.plot(y=0)
    plt.close()


def test_plot_bounds():
    _ = SCENE_FULL.plot(x=0, hlim=[-0.45, 0.45])
    plt.close()
    _ = SCENE_FULL.plot(x=0, vlim=[-0.45, 0.45])
    plt.close()
    _ = SCENE_FULL.plot(x=0, hlim=[-0.45, 0.45], vlim=[-0.45, 0.45])
    plt.close()


def test_structure_alpha():
    _ = SCENE_FULL.plot_structures_eps(x=0, alpha=None)
    plt.close()
    _ = SCENE_FULL.plot_structures_eps(x=0, alpha=-1)
    plt.close()
    _ = SCENE_FULL.plot_structures_eps(x=0, alpha=1)
    plt.close()
    _ = SCENE_FULL.plot_structures_eps(x=0, alpha=0.5)
    plt.close()
    _ = SCENE_FULL.plot_structures_eps(x=0, alpha=0.5, cbar=True)
    plt.close()
    new_structs = [
        td.Structure(geometry=s.geometry, medium=SCENE_FULL.medium) for s in SCENE_FULL.structures
    ]
    S2 = SCENE_FULL.copy(update=dict(structures=new_structs))
    _ = S2.plot_structures_eps(x=0, alpha=0.5)
    plt.close()


def test_filter_structures():
    s1 = td.Structure(geometry=td.Box(size=(1, 1, 1)), medium=SCENE.medium)
    s2 = td.Structure(geometry=td.Box(size=(1, 1, 1), center=(1, 1, 1)), medium=SCENE.medium)
    plane = td.Box(center=(0, 0, 1.5), size=(td.inf, td.inf, 0))
    SCENE._filter_structures_plane_medium(structures=[s1, s2], plane=plane)


def test_get_structure_plot_params():
    pp = SCENE_FULL._get_structure_plot_params(mat_index=0, medium=SCENE_FULL.medium)
    assert pp.facecolor == "white"
    pp = SCENE_FULL._get_structure_plot_params(mat_index=1, medium=td.PEC)
    assert pp.facecolor == "gold"
    pp = SCENE_FULL._get_structure_eps_plot_params(
        medium=SCENE_FULL.medium, freq=1, eps_min=1, eps_max=2
    )
    assert float(pp.facecolor) == 1.0
    pp = SCENE_FULL._get_structure_eps_plot_params(medium=td.PEC, freq=1, eps_min=1, eps_max=2)
    assert pp.facecolor == "gold"


def test_num_mediums():
    """Make sure we error if too many mediums supplied."""

    structures = []
    for i in range(MAX_NUM_MEDIUMS):
        structures.append(
            td.Structure(geometry=td.Box(size=(1, 1, 1)), medium=td.Medium(permittivity=i + 1))
        )
    _ = td.Scene(
        structures=structures,
    )

    with pytest.raises(pd.ValidationError):
        structures.append(
            td.Structure(geometry=td.Box(size=(1, 1, 1)), medium=td.Medium(permittivity=i + 2))
        )
        _ = td.Scene(structures=structures)


def _test_names_default():
    """makes sure default names are set"""

    scene = td.Scene(
        size=(2.0, 2.0, 2.0),
        structures=[
            td.Structure(
                geometry=td.Box(size=(1, 1, 1), center=(-1, 0, 0)),
                medium=td.Medium(permittivity=2.0),
            ),
            td.Structure(
                geometry=td.Box(size=(1, 1, 1), center=(0, 0, 0)),
                medium=td.Medium(permittivity=2.0),
            ),
            td.Structure(
                geometry=td.Sphere(radius=1.4, center=(1.0, 0.0, 1.0)), medium=td.Medium()
            ),
            td.Structure(
                geometry=td.Cylinder(radius=1.4, length=2.0, center=(1.0, 0.0, -1.0), axis=1),
                medium=td.Medium(),
            ),
        ],
    )

    for i, structure in enumerate(scene.structures):
        assert structure.name == f"structures[{i}]"


def test_names_unique():

    with pytest.raises(pd.ValidationError):
        _ = td.Scene(
            structures=[
                td.Structure(
                    geometry=td.Box(size=(1, 1, 1), center=(-1, 0, 0)),
                    medium=td.Medium(permittivity=2.0),
                    name="struct1",
                ),
                td.Structure(
                    geometry=td.Box(size=(1, 1, 1), center=(0, 0, 0)),
                    medium=td.Medium(permittivity=2.0),
                    name="struct1",
                ),
            ],
        )


def test_perturbed_mediums_copy():

    # Non-dispersive
    pp_real = td.ParameterPerturbation(
        heat=td.LinearHeatPerturbation(
            coeff=-0.01,
            temperature_ref=300,
            temperature_range=(200, 500),
        ),
    )

    pp_complex = td.ParameterPerturbation(
        heat=td.LinearHeatPerturbation(
            coeff=0.01j,
            temperature_ref=300,
            temperature_range=(200, 500),
        ),
        charge=td.LinearChargePerturbation(
            electron_coeff=-1e-21,
            electron_ref=0,
            electron_range=(0, 1e20),
            hole_coeff=-2e-21,
            hole_ref=0,
            hole_range=(0, 0.5e20),
        ),
    )

    coords = dict(x=[1, 2], y=[3, 4], z=[5, 6])
    temperature = td.SpatialDataArray(300 * np.ones((2, 2, 2)), coords=coords)
    electron_density = td.SpatialDataArray(1e18 * np.ones((2, 2, 2)), coords=coords)
    hole_density = td.SpatialDataArray(2e18 * np.ones((2, 2, 2)), coords=coords)

    pmed1 = td.PerturbationMedium(permittivity=3, permittivity_perturbation=pp_real)

    pmed2 = td.PerturbationPoleResidue(
        poles=[(1j, 3), (2j, 4)],
        poles_perturbation=[(None, pp_real), (pp_complex, None)],
    )

    struct = td.Structure(geometry=td.Box(center=(0, 0, 0), size=(1, 1, 1)), medium=pmed2)

    scene = td.Scene(
        medium=pmed1,
        structures=[struct],
    )

    # no perturbations provided -> regular mediums
    new_scene = scene.perturbed_mediums_copy()

    assert isinstance(new_scene.medium, td.Medium)
    assert isinstance(new_scene.structures[0].medium, td.PoleResidue)

    # perturbations provided -> custom mediums
    new_scene = scene.perturbed_mediums_copy(temperature)
    new_scene = scene.perturbed_mediums_copy(temperature, None, hole_density)
    new_scene = scene.perturbed_mediums_copy(temperature, electron_density, hole_density)

    assert isinstance(new_scene.medium, td.CustomMedium)
    assert isinstance(new_scene.structures[0].medium, td.CustomPoleResidue)


def test_max_geometry_validation():
    too_many = [td.Box(size=(1, 1, 1)) for _ in range(MAX_GEOMETRY_COUNT + 1)]

    fine = [
        td.Structure(
            geometry=td.ClipOperation(
                operation="union",
                geometry_a=td.Box(size=(1, 1, 1)),
                geometry_b=td.GeometryGroup(geometries=too_many),
            ),
            medium=td.Medium(permittivity=2.0),
        ),
        td.Structure(
            geometry=td.GeometryGroup(geometries=too_many),
            medium=td.Medium(permittivity=2.0),
        ),
    ]
    _ = td.Scene(structures=fine)

    not_fine = [
        td.Structure(
            geometry=td.ClipOperation(
                operation="difference",
                geometry_a=td.Box(size=(1, 1, 1)),
                geometry_b=td.GeometryGroup(geometries=too_many),
            ),
            medium=td.Medium(permittivity=2.0),
        ),
    ]
    with pytest.raises(pd.ValidationError, match=f" {MAX_GEOMETRY_COUNT + 2} "):
        _ = td.Scene(structures=not_fine)



import pytest
import pydantic.v1 as pd
import numpy as np
from matplotlib import pyplot as plt

import tidy3d as td

from tidy3d import FluidSpec, SolidSpec
from tidy3d import UniformHeatSource
from tidy3d import (
    TemperatureBC,
    HeatFluxBC,
    ConvectionBC,
    HeatBoundarySpec,
)
from tidy3d import (
    StructureBoundary,
    StructureStructureInterface,
    SimulationBoundary,
    StructureSimulationBoundary,
    MediumMediumInterface,
)
from tidy3d import UniformUnstructuredGrid, DistanceUnstructuredGrid
from tidy3d import HeatSimulation
from tidy3d import HeatSimulationData
from tidy3d import TemperatureMonitor
from tidy3d import TemperatureData

from ..utils import STL_GEO, assert_log_level, log_capture


def make_heat_mediums():
    fluid_medium = td.Medium(
        permittivity=3,
        heat_spec=FluidSpec(),
        name="fluid_medium",
    )
    solid_medium = td.Medium(
        permittivity=5,
        conductivity=0.01,
        heat_spec=SolidSpec(
            capacity=2,
            conductivity=3,
        ),
        name="solid_medium",
    )

    return fluid_medium, solid_medium


def test_heat_medium():
    _, solid_medium = make_heat_mediums()

    with pytest.raises(pd.ValidationError):
        _ = solid_medium.heat_spec.updated_copy(capacity=-1)

    with pytest.raises(pd.ValidationError):
        _ = solid_medium.heat_spec.updated_copy(conductivity=-1)


def make_heat_structures():
    fluid_medium, solid_medium = make_heat_mediums()

    box = td.Box(center=(0, 0, 0), size=(1, 1, 1))

    fluid_structure = td.Structure(
        geometry=box,
        medium=fluid_medium,
        name="fluid_structure",
    )

    solid_structure = td.Structure(
        geometry=box.updated_copy(center=(1, 1, 1)),
        medium=solid_medium,
        name="solid_structure",
    )

    return fluid_structure, solid_structure


def test_heat_structures():
    _, _ = make_heat_structures()


def make_heat_bcs():
    bc_temp = TemperatureBC(temperature=300)
    bc_flux = HeatFluxBC(flux=20)
    bc_conv = ConvectionBC(ambient_temperature=400, transfer_coeff=0.2)

    return bc_temp, bc_flux, bc_conv


def test_heat_bcs():
    bc_temp, bc_flux, bc_conv = make_heat_bcs()

    with pytest.raises(pd.ValidationError):
        _ = TemperatureBC(temperature=-10)

    with pytest.raises(pd.ValidationError):
        _ = ConvectionBC(ambient_temperature=-400, transfer_coeff=0.2)

    with pytest.raises(pd.ValidationError):
        _ = ConvectionBC(ambient_temperature=400, transfer_coeff=-0.2)


def make_heat_mnt():
    return TemperatureMonitor(size=(1, 2, 3), name="test")


def test_heat_mnt():
    temp_mnt = make_heat_mnt()

    with pytest.raises(pd.ValidationError):
        _ = temp_mnt.updated_copy(name=None)

    with pytest.raises(pd.ValidationError):
        _ = temp_mnt.updated_copy(size=(-1, 2, 3))


def make_heat_mnt_data():
    temp_mnt = make_heat_mnt()

    nx, ny, nz = 9, 6, 5
    x = np.linspace(-1, 1, nx)
    y = np.linspace(-2, 2, ny)
    z = np.linspace(-3, 3, nz)
    T = np.random.default_rng().uniform(300, 350, (nx, ny, nz))
    coords = dict(x=x, y=y, z=z)
    temperature_field = td.SpatialDataArray(T, coords=coords)

    return TemperatureData(monitor=temp_mnt, temperature=temperature_field)


def test_heat_mnt_data():
    _ = make_heat_mnt_data


def make_uniform_grid_spec():
    return UniformUnstructuredGrid(dl=0.1, min_edges_per_circumference=5, min_edges_per_side=3)


def make_distance_grid_spec():
    return DistanceUnstructuredGrid(
        dl_interface=0.1, dl_bulk=1, distance_interface=1, distance_bulk=2
    )


def test_grid_spec():
    grid_spec = make_uniform_grid_spec()
    with pytest.raises(pd.ValidationError):
        _ = grid_spec.updated_copy(dl=0)
    with pytest.raises(pd.ValidationError):
        _ = grid_spec.updated_copy(min_edges_per_circumference=-1)
    with pytest.raises(pd.ValidationError):
        _ = grid_spec.updated_copy(min_edges_per_side=-1)

    grid_spec = make_distance_grid_spec()
    with pytest.raises(pd.ValidationError):
        grid_spec.updated_copy(dl_interface=-1)
    with pytest.raises(pd.ValidationError):
        grid_spec.updated_copy(distance_interface=2, distance_bulk=1)


def make_heat_source():
    return UniformHeatSource(structures=["solid_structure"], rate=100)


def test_heat_source():
    _ = make_heat_source()


def make_heat_sim():
    fluid_medium, solid_medium = make_heat_mediums()
    fluid_structure, solid_structure = make_heat_structures()
    bc_temp, bc_flux, bc_conv = make_heat_bcs()
    heat_source = make_heat_source()

    pl1 = HeatBoundarySpec(
        condition=bc_conv, placement=MediumMediumInterface(mediums=["fluid_medium", "solid_medium"])
    )
    pl2 = HeatBoundarySpec(
        condition=bc_flux, placement=StructureBoundary(structure="solid_structure")
    )
    pl3 = HeatBoundarySpec(
        condition=bc_flux,
        placement=StructureStructureInterface(structures=["fluid_structure", "solid_structure"]),
    )
    pl4 = HeatBoundarySpec(condition=bc_temp, placement=SimulationBoundary())
    pl5 = HeatBoundarySpec(
        condition=bc_temp, placement=StructureSimulationBoundary(structure="fluid_structure")
    )

    grid_spec = make_uniform_grid_spec()

    temp_mnt = make_heat_mnt()

    heat_sim = HeatSimulation(
        medium=fluid_medium,
        structures=[fluid_structure, solid_structure],
        center=(0, 0, 0),
        size=(2, 2, 2),
        boundary_spec=[pl1, pl2, pl3, pl4, pl5],
        grid_spec=grid_spec,
        sources=[heat_source],
        monitors=[temp_mnt],
    )

    return heat_sim


def test_heat_sim():
    bc_temp, bc_flux, bc_conv = make_heat_bcs()
    heat_sim = make_heat_sim()

    _ = heat_sim.plot(x=0)

    # wrong names given
    for pl in [
        HeatBoundarySpec(
            condition=bc_temp, placement=MediumMediumInterface(mediums=["badname", "fluid_medium"])
        ),
        HeatBoundarySpec(condition=bc_flux, placement=StructureBoundary(structure="no_box")),
        HeatBoundarySpec(
            condition=bc_conv,
            placement=StructureStructureInterface(structures=["no_box", "solid_structure"]),
        ),
        HeatBoundarySpec(
            condition=bc_temp, placement=StructureSimulationBoundary(structure="no_mesh")
        ),
    ]:
        with pytest.raises(pd.ValidationError):
            _ = heat_sim.updated_copy(boundary_spec=[pl])

    with pytest.raises(pd.ValidationError):
        _ = heat_sim.updated_copy(sources=[UniformHeatSource(structures=["noname"])], rate=-10)

    # polyslab support
    vertices = np.array([(0, 0), (1, 0), (1, 1)])
    p = td.PolySlab(vertices=vertices, axis=2, slab_bounds=(-1, 1))
    _, structure = make_heat_structures()
    structure = structure.updated_copy(geometry=p, name="polyslab")
    _ = heat_sim.updated_copy(structures=list(heat_sim.structures) + [structure])

    # test unsupported yet geometries
    structure = structure.updated_copy(geometry=STL_GEO, name="stl")
    with pytest.raises(pd.ValidationError):
        _ = heat_sim.updated_copy(structures=list(heat_sim.structures) + [structure])

    # test unsupported yet zero dimension domains
    with pytest.raises(pd.ValidationError):
        _ = heat_sim.updated_copy(center=(0, 0, 0), size=(0, 2, 2))

    with pytest.raises(pd.ValidationError):
        _ = heat_sim.updated_copy(center=(0, 0, 0), size=(1, 0, 0))

    temp_mnt = make_heat_mnt()

    with pytest.raises(pd.ValidationError):
        heat_sim.updated_copy(monitors=[temp_mnt, temp_mnt])

    _ = heat_sim.plot(x=0)
    plt.close()

    _ = heat_sim.plot_heat_conductivity(y=0)
    plt.close()

    heat_sim = heat_sim.updated_copy(symmetry=(0, 1, 1))
    _ = heat_sim.plot_heat_conductivity(z=0, colorbar="source")
    plt.close()

    with pytest.raises(pd.ValidationError):
        _ = heat_sim.updated_copy(symmetry=(-1, 0, 1))


@pytest.mark.parametrize("shift_amount, log_level", ((1, None), (2, "WARNING")))
def test_heat_sim_bounds(shift_amount, log_level, log_capture):
    """make sure bounds are working correctly"""

    # make sure all things are shifted to this central location
    CENTER_SHIFT = (-1.0, 1.0, 100.0)

    def place_box(center_offset):

        shifted_center = tuple(c + s for (c, s) in zip(center_offset, CENTER_SHIFT))

        _ = td.HeatSimulation(
            size=(1.5, 1.5, 1.5),
            center=CENTER_SHIFT,
            structures=[
                td.Structure(
                    geometry=td.Box(size=(1, 1, 1), center=shifted_center), medium=td.Medium()
                )
            ],
            grid_spec=td.UniformUnstructuredGrid(dl=0.1),
        )

    # create all permutations of squares being shifted 1, -1, or zero in all three directions
    bin_strings = [list(format(i, "03b")) for i in range(8)]
    bin_ints = [[int(b) for b in bin_string] for bin_string in bin_strings]
    bin_ints = np.array(bin_ints)
    bin_signs = 2 * (bin_ints - 0.5)

    # test all cases where box is shifted +/- 1 in x,y,z and still intersects
    for amp in bin_ints:
        for sign in bin_signs:
            center = shift_amount * amp * sign
            if np.sum(center) < 1e-12:
                continue
            place_box(tuple(center))
    assert_log_level(log_capture, log_level)


@pytest.mark.parametrize(
    "box_size,log_level",
    [
        ((1, 0.1, 0.1), "WARNING"),
        ((0.1, 1, 0.1), "WARNING"),
        ((0.1, 0.1, 1), "WARNING"),
    ],
)
def test_sim_structure_extent(log_capture, box_size, log_level):
    """Make sure we warn if structure extends exactly to simulation edges."""

    box = td.Structure(geometry=td.Box(size=box_size), medium=td.Medium(permittivity=2))
    _ = td.HeatSimulation(
        size=(1, 1, 1),
        structures=[box],
        grid_spec=td.UniformUnstructuredGrid(dl=0.1),
    )

    assert_log_level(log_capture, log_level)


def make_heat_sim_data():
    heat_sim = make_heat_sim()
    temp_data = make_heat_mnt_data()

    heat_sim_data = HeatSimulationData(
        simulation=heat_sim,
        data=[temp_data],
    )

    return heat_sim_data


def test_sim_data():
    heat_sim_data = make_heat_sim_data()
    _ = heat_sim_data.plot_field("test", z=0)
    plt.close()

    with pytest.raises(KeyError):
        _ = heat_sim_data.plot_field("test3", x=0)

    with pytest.raises(pd.ValidationError):
        _ = heat_sim_data.updated_copy(data=[heat_sim_data.data[0]] * 2)

    temp_mnt = make_heat_mnt()
    temp_mnt = temp_mnt.updated_copy(name="test2")

    sim = heat_sim_data.simulation.updated_copy(monitors=[temp_mnt])

    with pytest.raises(pd.ValidationError):
        _ = heat_sim_data.updated_copy(simulation=sim)



import pytest
import gdstk
import numpy as np
import tidy3d as td
import pydantic.v1 as pd


def test_to_gds(tmp_path):
    geometry = td.Box(size=(2, 2, 2))
    medium = td.Medium()
    structure = td.Structure(geometry=geometry, medium=medium)

    fname = str(tmp_path / "structure.gds")
    structure.to_gds_file(fname, x=0, gds_cell_name="X")
    cell = gdstk.read_gds(fname).cells[0]
    assert cell.name == "X"
    assert len(cell.polygons) == 1
    assert np.allclose(cell.polygons[0].area(), 4.0)


def test_custom_medium_to_gds(tmp_path):
    geometry = td.Box(size=(2, 2, 2))

    nx, ny, nz = 100, 90, 80
    x = np.linspace(0, 2, nx)
    y = np.linspace(-1, 1, ny)
    z = np.linspace(-1, 1, nz)
    f = np.array([td.C_0])
    mx, my, mz, _ = np.meshgrid(x, y, z, f, indexing="ij", sparse=True)
    data = 1 + 1 / (1 + (mx - 1) ** 2 + my**2 + mz**2)
    eps_diagonal_data = td.ScalarFieldDataArray(data, coords=dict(x=x, y=y, z=z, f=f))
    eps_components = {f"eps_{d}{d}": eps_diagonal_data for d in "xyz"}
    eps_dataset = td.PermittivityDataset(**eps_components)
    medium = td.CustomMedium(eps_dataset=eps_dataset, name="my_medium")
    structure = td.Structure(geometry=geometry, medium=medium)

    fname = str(tmp_path / "structure-custom-x.gds")
    structure.to_gds_file(fname, x=1, permittivity_threshold=1.5, frequency=td.C_0)
    cell = gdstk.read_gds(fname).cells[0]
    assert np.allclose(cell.area(), np.pi, atol=1e-2)

    fname = str(tmp_path / "structure-custom-z.gds")
    structure.to_gds_file(fname, z=0, permittivity_threshold=1.5, frequency=td.C_0)
    cell = gdstk.read_gds(fname).cells[0]
    assert np.allclose(cell.area(), np.pi / 2, atol=3e-2)

    fname = str(tmp_path / "structure-empty.gds")
    structure.to_gds_file(fname, x=-0.1, permittivity_threshold=1.5, frequency=td.C_0)
    cell = gdstk.read_gds(fname).cells[0]
    assert len(cell.polygons) == 0


def test_non_symmetric_custom_medium_to_gds(tmp_path):
    geometry = td.Box(size=(1, 2, 1), center=(0.5, 0, 2.5))

    nx, ny, nz = 150, 80, 180
    x = np.linspace(0, 2, nx)
    y = np.linspace(-1, 1, ny)
    z = np.linspace(2, 3, nz)
    f = np.array([td.C_0])
    mx, my, mz, _ = np.meshgrid(x, y, z, f, indexing="ij", sparse=True)
    data = 1 + mx + 0 * my + (mz - 2) ** 2
    print(data.min(), data.max())

    eps_diagonal_data = td.ScalarFieldDataArray(data, coords=dict(x=x, y=y, z=z, f=f))
    eps_components = {f"eps_{d}{d}": eps_diagonal_data for d in "xyz"}
    eps_dataset = td.PermittivityDataset(**eps_components)
    medium = td.CustomMedium(eps_dataset=eps_dataset, name="my_medium")
    structure = td.Structure(geometry=geometry, medium=medium)

    fname = str(tmp_path / "structure-non-symmetric.gds")
    structure.to_gds_file(fname, y=0, permittivity_threshold=2.0, frequency=td.C_0)
    cell = gdstk.read_gds(fname).cells[0]
    assert np.allclose(cell.bounding_box(), ((0, 2), (1, 3)), atol=0.1)
    assert gdstk.inside([(0.1, 2.1), (0.5, 2.5), (0.9, 2.9)], cell.polygons) == (False, False, True)


@pytest.mark.parametrize("axis", [0, 1, 2])
def test_invalid_polyslab(axis):
    medium = td.Medium()
    vertices = [(-1, -2), (-1, 1), (1, 2), (2, 1), (0, 1), (0, 0), (1.5, -0.5), (0, -1), (0, -2)]
    i = (axis + 1) % 3
    j = (axis + 2) % 3

    ps = td.PolySlab(vertices=vertices, slab_bounds=(-1, 1), sidewall_angle=0.1, axis=axis)
    box = td.Box(size=(1, 1, 1))

    geo0 = ps.rotated(-np.pi / 3, axis)
    _ = td.Structure(geometry=geo0, medium=medium)

    geo1 = ps.rotated(-np.pi / 3, i).scaled(2, 2, 2).translated(-1, 0.5, 2).rotated(np.pi / 3, i)
    _ = td.Structure(geometry=geo1, medium=medium)

    geo2 = ps.rotated(np.pi / 4, i)
    with pytest.raises(pd.ValidationError):
        _ = td.Structure(geometry=geo2, medium=medium)

    geo3 = ps.rotated(np.pi / 5, j)
    with pytest.raises(pd.ValidationError):
        _ = td.Structure(geometry=geo3, medium=medium)

    geo4 = ps.rotated(np.pi / 6, (1, 1, 1))
    with pytest.raises(pd.ValidationError):
        _ = td.Structure(geometry=geo4, medium=medium)

    geo5 = td.GeometryGroup(geometries=[ps]).rotated(np.pi / 2, j)
    with pytest.raises(pd.ValidationError):
        _ = td.Structure(geometry=geo5, medium=medium)

    geo6 = td.GeometryGroup(geometries=[ps - box]).rotated(np.pi / 2, i)
    with pytest.raises(pd.ValidationError):
        _ = td.Structure(geometry=geo6, medium=medium)

    geo7 = td.GeometryGroup(geometries=[(ps - box).rotated(np.pi / 4, j)]).rotated(-np.pi / 4, j)
    _ = td.Structure(geometry=geo7, medium=medium)



"""Custom Tidy3D exceptions"""
from .core_config import get_logger


class WebError(Exception):
    """Any error in tidy3d"""

    def __init__(self, message: str = None):
        """Log just the error message and then raise the Exception."""
        log = get_logger()
        super().__init__(message)
        log.error(message)
